/*
 * Copyright (C) 2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.Reader;
import java.text.MessageFormat;
import java.util.List;
import java.util.ResourceBundle;

import net.siefkes.nlstego.textgen.StegoDecoder;
import net.siefkes.nlstego.textgen.StegoEncoder;
import net.siefkes.nlstego.textgen.TextModel;
import net.siefkes.nlstego.util.InOutUtils;
import net.siefkes.nlstego.util.Util;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.WordUtils;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * The main entry point of this program.
 * 
 * @author Christian Siefkes
 * @version $Revision: 1.13 $, $Date: 2005/07/12 17:02:07 $, $Author: siefkes $
 */
public class NLStego {

    /**
     * The ID of this software, used as base name of the config file etc.:
     * {@value}.
     */
    public static final String ID = "nlstego";

    /**
     * The name of this software: {@value}.
     */
    public static final String NAME = "NLStego";

    /**
     * The current version number of this software: {@value}.
     */
    public static final String VERSION_NO = "0.9-dev";

    /**
     * Marker used for invalid or missing commands.
     */
    public static final char INVALID_COMMAND = '0';

    /**
     * The logger used in this system. All log messages are written to stderr
     * (the standard error stream).
     */
    public static final Log LOG = LogFactory.getLog(NLStego.class);

    /**
     * Messages used by this program.
     */
    public static final ResourceBundle RES =
        ResourceBundle.getBundle("net.siefkes.nlstego.Resources");


    /**
     * Static main method: create and configure a new instance and
     * {@linkplain #execute(char, List) executes} the given command
     * (first argument) on the given arguments.
     *
     * @param args the command-line parameters: first element must be the
     * command to execute (<code>e</code> to encode, <code>d</code> to decode,
     * or <code>g</code> to generate random text); elements starting with  "-"
     * are processed as configuration options (cf.
     * {@link Util#configureFromArgs(Configuration, String[])})
     */
    public static void main(final String[] args) {
        try {
            // modify standard configuration + collect remaining arguments
            final Configuration config;
            final List<String> arguments;

            try {
                config  = Util.loadConfiguration(ID);
                arguments = Util.configureFromArgs(config, args);
            } catch (Exception e) {
                LOG.fatal(RES.getString("illegal-option.error"), e);
                return;
            }

            // determine the command to use
            final char command;

            if ((arguments.size()) > 0) {
                final String firstArg = arguments.get(0);
                arguments.remove(0);

                // command have exactly one letter
                if (firstArg.length() == 1) {
                    command = firstArg.charAt(0);
                } else {
                    command = INVALID_COMMAND;
                }
            } else {
                // missing command
                command = INVALID_COMMAND;
            }

            // create and execute a new instance
            final NLStego main = new NLStego(config);
            main.execute(command, arguments);
        } catch (Throwable t) {
            LOG.fatal(RES.getString("exit.fatal"), t);
        }
    }


    /**
     * Used to configure this instance.
     */
    private final Configuration config;

    /**
     * The stream to use for error and warn messages.
     */
    private final PrintStream err;

    /**
     * The stream to use for input.
     */
    private final InputStream in;

    /**
     * The stream to use for output.
     */
    private final PrintStream out;


    /**
     * Creates a new instance, using the standard input/output/error streams
     * for I/O.
     *
     * @param myConfig used to configure this instance
     */
    public NLStego(final Configuration myConfig) {
        this(myConfig, System.in, System.out, System.err);
    }

    /**
     * Creates a new instance.
     *
     * @param myConfig used to configure this instance
     * @param myInStream stream used to read input
     * @param myOutStream stream used to write output
     * @param myErrStream stream used to write error and warn message
     */
    public NLStego(final Configuration myConfig, final InputStream myInStream,
            final PrintStream myOutStream, final PrintStream myErrStream) {
        super();
        config = myConfig;
        in = myInStream;
        out = myOutStream;
        err = myErrStream;
    }


    /**
     * Build a text model from a set of texts.
     *
     * @param filePaths array of paths of the text to use for building the
     * model; resolved via the
     * <a href="http://jakarta.apache.org/commons/vfs/">Commons VFS</a>
     * @param recurse whether to recursively add the children of folders and
     * other files containing nested entries
     * @return the created text model
     * @throws IOException if an error occurs while reading training texts
     */
    public TextModel buildTextModel(final String[] filePaths,
            final boolean recurse) throws IOException {
        LOG.info(RES.getString("building-model.info"));
        final TextModel textModel = new TextModel(config);
        textModel.train(filePaths, recurse);
        return textModel;
    }

    /**
     * Executes the given command on the given arguments (if any). Also prints
     * help and/or version info if requested ("-h"/"-v" options).
     *
     * @param command the command to execute
     * @param arguments a list of any further arguments for this command
     * @throws IOException if an I/O error occurs
     */
    public void execute(final char command, final List<String> arguments)
    throws IOException {
        final boolean printedVersion;
        if (config.getBoolean("v", false)) {
            // "-v" option specified: print version info
            printVersion();
            printedVersion = true;

            // exit if there is nothing else to do
            if (arguments.isEmpty()) {
                return;
            }
        } else {
            printedVersion = false;
        }

        if ((command == INVALID_COMMAND) || config.getBoolean("h", false)) {
            // invalid/missing command or "-h" specified: print
            // version info + short help message to standard out and exit
            if (!printedVersion) { // don't print twice
                printVersion();
            }

            printUsage();
            return;
        }

        // check whether to use compression when encoding/decoding
        final boolean compress = config.getBoolean("c", false);

        // build text model
        final String[] modelPaths = config.getStringArray("m");
        if (modelPaths.length == 0) {
            err.println(RES.getString("missing-m.error"));
            printUsage();
            return;
        }
        final TextModel textModel = buildTextModel(modelPaths, true);
        byte[] bitSequence;

        // execute specified command
        switch (command) {
        case 'd':
            // decode bit sequence
            final StegoDecoder decoder = new StegoDecoder(config, textModel);
            readingInputMsg();
            bitSequence = decoder.decode(new InputStreamReader(in));
            if (LOG.isDebugEnabled()) {
                LOG.debug("Decoded bit sequence (length " + bitSequence.length
                        + "): " + ArrayUtils.toString(bitSequence));
            }

            // write result, uncompressing it if configured
            if (compress) {
                out.write(InOutUtils.uncompress(bitSequence));
            } else {
                out.write(bitSequence);
            }
            break;

        case 'e':
            // encode bit sequence
            final StegoEncoder encoder = new StegoEncoder(config, textModel);
            readingInputMsg();
            bitSequence = IOUtils.toByteArray(in);
            if (compress) {
                bitSequence = InOutUtils.compress(bitSequence);
            }
            if (LOG.isDebugEnabled()) {
                LOG.debug("Bit sequence to encode (length " + bitSequence.length
                        + "): " + ArrayUtils.toString(bitSequence));
            }
            printWrapped(encoder.encode(bitSequence));
            break;

        case 'g':
            // generate random text
            final int numWords;
            if (arguments.isEmpty()) {
                err.println(RES.getString("missing-numWords.error"));
                printUsage();
                return;
            } else {
                numWords = Util.asInt(arguments.get(0));
            }

            printWrapped(generateText(textModel, numWords));
            break;

        default:
            LOG.fatal(MessageFormat.format(
                    RES.getString("unknown-command.error"), command));
        }
    }

    /**
     * Generates random text.
     *
     * @param textModel the text model to use
     * @param numWords the minimum number of words to generate
     * @return the generated text
     * @throws IOException if an I/O error occurs
     */
    public String generateText(final TextModel textModel, final int numWords)
    throws IOException {
        LOG.info(RES.getString("generating-text.info"));
        final StringBuilder appender = new StringBuilder();
        textModel.predictTokens(numWords, true, appender);
        return appender.toString();
    }

    /**
     * Helper method that prints a text to the output stream, adding
     * linebreaks after 75 characters. A final newline will be appended and
     * the output stream will be flushed after writing the wrapped text.
     *
     * @param text the text to print
     */
    public void printWrapped(final String text) {
        printWrapped(text, 75);
    }

    /**
     * Helper method that prints a text to the output stream, adding
     * linebreaks after the specified number of characters. A final newline will
     * be appended and the output stream will be flushed after writing the
     * wrapped text.
     *
     * @param n the maximum number of characters in a line
     * @param text the text to print
     */
    public void printWrapped(final String text, final int n) {
        out.println(WordUtils.wrap(text, n));
        out.flush();
    }

    /**
     * Helper method that prints usage instructions to the error stream.
     */
    public void printUsage() {
        err.println();
        // load Usage.txt from classpath
        final ClassLoader classLoader = this.getClass().getClassLoader();
        final InputStream stream = classLoader.getResourceAsStream("Usage.txt");
        try {
            final Reader reader = new InputStreamReader(stream,
                    InOutUtils.STANDARD_UNICODE_CHARSET);
            try {
                IOUtils.copy(reader, err);
            } finally {
                InOutUtils.tryToClose(reader);
            }
        } catch (IOException ioe) {
            LOG.error("Error while printing usage notes", ioe);
        }
    }

    /**
     * Helper method that prints name and version number of this program to the
     * error stream.
     */
    public void printVersion() {
        err.println(NAME + ' ' + VERSION_NO);
    }

    /**
     * Notify user that we are reading input.
     */
    private void readingInputMsg() {
        LOG.info(RES.getString("reading-input.info"));
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("version", VERSION_NO)
            .toString();
    }

}
