/*
 * Copyright (C) 2003-2005 Christian Siefkes <christian@siefkes.net>
 * Parts of this code are Copyright (C) 2001-2002 by MUNDWERK AG.
 * Re-released under GNU GPL with permission of MUNDWERK AG.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.predict;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.SortedMap;

import org.apache.commons.configuration.Configuration;

/**
 * Abstract class for collecting statistics about the fetching of items
 * that can be used for predicting. <p>
 * Created (11.09.2001 16:35:54).<p>
 *
 * @author Christian Siefkes
 * @version $Revision: 1.8 $, $Date: 2005/07/28 15:03:09 $, $Author: siefkes $
 */
public abstract class FetchStats implements Serializable {

    /**
     * Constant used to mark the start of a session. <p>
     * Created (27.09.2001 12:11:47) by Christian Siefkes.<p>
     */
    protected static final Integer START_OF_SESSION = Integer.valueOf(-1);

    /**
     * Constant used to mark the end of a session. <p>
     * Created (01.10.2001 10:53:53) by Christian Siefkes.<p>
     */
    protected static final Integer END_OF_SESSION = Integer.valueOf(-2);

    /**
     * Used to manage the IDs of all instances--contains the next ID
     * that isn't yet used. The constructor of a new instance stores the
     * current value as instance ID and then increments this field.
     */
    private static int nextAvailableID = 0;

    /**
     * Used to synchronize the {@link #nextAvailableID} field.
     */
    private static final Object ID_GUARD = new Object();

    /**
     * The unique ID of this instance.
     */
    private final int id;

    /**
     * The unique ID of this instance as an Integer object, for easy storage
     * in collections etc.
     */
    private final Integer idInteger;

    /**
     * Optional filter that decides which predictions should be stored
     * in the stateless (global) model; if <code>null</code>, all predictions
     * are stored.
     */
    private final StatelessFilter statelessFilter;

    /**
     * Used by <code>unmapKey</code> to retrieve the keys represented
     * by integers: The n-th element of this ArrayList contains the key
     * String represented by n. <p>
     * Created (27.09.2001 15:46:01) by Christian Siefkes.<p>
     */
    private static final ArrayList<String> MAPPINGS_LIST =
        new ArrayList<String>();

    /**
     * Used by <code>mapKey</code> to store the integer representations
     * of keys (using key Strings as keys and FetchStatsKey's
     * as values). <p>
     * Created (27.09.2001 15:46:04) by Christian Siefkes.<p>
     */
    private static final HashMap<String, FetchStatsKey> MAPPINGS_MAP =
        new HashMap<String, FetchStatsKey>();

/**
 * Creates a new instance. <p>
 * Created (11.09.2001 16:35:54) by Christian Siefkes.<p>
 *
 * @param conf the <code>Configuration</code> to use
 * @param filter optional filter that decides which predictions should be stored
 * in the stateless (global) model; if <code>null</code>, all predictions are
 * stored
 */
protected FetchStats(final Configuration conf, final StatelessFilter filter) {
    super();
    statelessFilter = filter;

    // determine ID, afterwards incrementing static ID counter
    synchronized (ID_GUARD) {
        id = nextAvailableID++;
        idInteger = Integer.valueOf(id);
    }

} // method

/**
 * Adds a request of an item, training the model. <p>
 * Created (27.09.2001 12:16:43) by Christian Siefkes.<p>
 *
 * @param key the key of the item currently requested
 * fetching the last document in this session
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 */
public final void add(final String key, final int sessionID) {
    add(key, sessionID, true);
}

/**
 * Adds a request of an item. <p>
 * Created (27.09.2001 12:16:43) by Christian Siefkes.<p>
 *
 * @param key the key of the item currently requested
 * fetching the last document in this session
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 * @param train <code>true</code> iff the model should be trained, i.e.
 * this request should be incorporated into the prediction model; otherwise
 * only the internal state will be changed so the next call to {@link
 * #predict} will find suitable predictions (this happens in any case)
 */
public abstract void add(final String key, final int sessionID,
    final boolean train);

/**
 * Maps the key used to identify a document to a compact
 * representation used for storing it in the FetchStats. <p>
 * Created (27.09.2001 15:20:15) by Christian Siefkes.<p>
 *
 * @param key the key of the document (the URL's external form)
 * @return a compact and unique representation of this key
 */
public static synchronized Integer mapKey(final String key) {
    FetchStatsKey fsKey;
    int representation;

    // is this key already mapped?
    fsKey = MAPPINGS_MAP.get(key);

    if (fsKey == null) {
        // no, we have to map it--it gets the next free value
        representation = MAPPINGS_LIST.size();
        fsKey = new FetchStatsKey(representation);

        // adding it at the end of the ArrayList
        MAPPINGS_LIST.add(key);

        // storing the resulting value in the HashMap
        MAPPINGS_MAP.put(key, fsKey);

//      debugOrInfo(NAME, "mapKey", "Mapped " + key + " to " + representation);
    } else {
        representation = fsKey.getRepresentation();
    } // else

    return Integer.valueOf(representation);
} // method

/**
 * Predicts the documents that will be probably fetched next. <p>
 * Created (27.09.2001 12:16:43) by Christian Siefkes.<p>
 *
 * @param number the number of results to return, >0
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 * @return a SortedMap containing mapped keys as keys and FetchStatsResult's as
 * values. The length of the map is max (<code>number</code>, all matching
 * documents that had at least <code>minAccesses</code>)
 */
public abstract SortedMap predict(final int number, final int sessionID);

/**
 * Marks a session as ended. <p>
 * Created (27.09.2001 12:16:43) by Christian Siefkes.<p>
 *
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 */
public abstract void sessionEnded(final int sessionID);

/**
 * Shuts down this entity. The intention is to free all resources occupied by
 * this entity and shut down all other entities that belong to this one. <p>
 * Created (04.03.2002 11:31:28) by Christian Siefkes.<p>
 *
 * @param requestgraceful a <code>boolean</code>; if set to true the entity
 * should wait for its current activity to stop and then shutdown, if set to
 * false it shuts down immediately
 */
public void shutdown(final boolean requestgraceful) {
    // currently, this method does nothing
} // method

/**
 * Prints statistical information about this model. This method might
 * take some time while the data is gathered.
 *
 * @return a string listing statistical details about this model
 */
public abstract String printModelStats();

    /**
     * Maps the key used to identify an item to a compact
     * representation used for storing it in the FetchStats. <p>
     * Created (27.09.2001 15:20:15) by Christian Siefkes.<p>
     *
     * @param representation the compact integer representation of a
     * key which must have been mapped before by <code>mapKey</code>
     * @return the represented key
     * @throws IndexOutOfBoundsException if the key wasn't mapped before
     */
    public static String unmapKey(final int representation)
            throws IndexOutOfBoundsException {
        String result;

        // retrieve the represented String
        // this might throw the IndexOutOfBoundsException
        result = MAPPINGS_LIST.get(representation);

        return result;
    } // method

    /**
     * Helper method that unmaps all keys from a collection.
     *
     * @param coll the collection to unmap
     * @return a list of {@linkplain #unmapKey(int) unmapped} keys
     */
    static List<String> unmapAllKeys(final Collection<Integer> coll) {
        final List<String> result = new ArrayList<String>(coll.size());
        final Iterator<Integer> collIter = coll.iterator();

        while (collIter.hasNext()) {
            result.add(unmapKey(collIter.next()));
        }
        return result;
    }

    /**
     * Helper method that unmaps all keys from a collection of
     * {@link FetchStatsResult}s.
     *
     * @param coll the collection to unmap
     * @return a list of {@linkplain #unmapKey(int) unmapped} keys
     */
    static List<String> unmapAllResults(
            final Collection<FetchStatsResult> coll) {
        final List<String> result = new ArrayList<String>(coll.size());
        final Iterator<FetchStatsResult> collIter = coll.iterator();

        while (collIter.hasNext()) {
            result.add(unmapKey(collIter.next().getMappedKey()));
        }
        return result;
    }

    /**
     * Returns the unique ID of this instance.
     *
     * @return the ID
     */
    public final int getID() {
        return id;
    }
    /**
     * Returns the unique ID of this instance as an Integer object, for easy
     * storage in collections etc..
     *
     * @return the ID Integer object
     */
    public final Integer getIDInteger() {
        return idInteger;
    }

    /**
     * Returns the optional filter that decides which predictions should be 
     * stored in the stateless (global) model; if <code>null</code>, all
     * predictions are stored.
     *
     * @return the value of the attribute
     */
    public StatelessFilter getStatelessFilter() {
        return statelessFilter;
    }

} // class
