/*
 * Copyright (C) 2003-2005 Christian Siefkes <christian@siefkes.net>
 * Parts of this code are Copyright (C) 2001-2002 by MUNDWERK AG.
 * Re-released under GNU GPL with permission of MUNDWERK AG.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.predict;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.siefkes.nlstego.util.InOutUtils;

import org.apache.commons.configuration.Configuration;


/**
 * Implementations of this abstract class handle performance measuring and
 * logging.
 * All instances of subclasses are registered and their status is printed
 * in regular intervals. To support this, each subclass must provide
 * appropriate <code>printMetrics</code> and <code>printCommon</code> methods.
 * Other methods of subclasses should call <code>checkPublication</code>
 * to ensure that publication takes place as scheduled. <p>
 * Created (02.01.2002 16:00:47).<p>
 *
 * @author Christian Siefkes
 * @version $Revision: 1.7 $, $Date: 2005/07/12 17:02:17 $, $Author: siefkes $
 */
public abstract class Metrics {

    /**
     * The number of digits used for formatting potentially small data,
     * where more digits than usually should be shown. <p>
     * Created (21.01.2002 14:57:57) by Christian Siefkes.<p>
     */
    protected static final int MORE_DIGITS = 5;

    /**
     * Contains all instances of subclasses. Also used as guard
     * for class-wide synchronization. <p>
     * Created (19.12.2001 11:40:57) by Christian Siefkes.<p>
     */
    private static final List<Metrics> ALL_METRICS = new ArrayList<Metrics>();

    /**
     * The used <code>Configuration</code>. <p>
     * Created (10.01.2002 13:17:50) by Christian Siefkes.<p>
     */
    private static Configuration config;

/**
 * Creates a new instance. <p>
 * Created (13.12.2001 10:45:58) by Christian Siefkes.<p>
 *
 * @param conf the <code>Configuration</code> to use
 */
Metrics(final Configuration conf) {
    super();

    // class-wide synchronization
    synchronized (ALL_METRICS) {
        // the first instance configures the static fields
        if (ALL_METRICS.isEmpty()) {
            setConfig(conf);
        } // if

        // register this instance
        ALL_METRICS.add(this);

    } // synchronized
} // method

/**
 * Collects the results of all metrics for publication. <p>
 * Created (15.01.2002 12:36:23) by Christian Siefkes.<p>
 *
 * @return the results of all metrics
 */
public static StringBuilder collectResults() {
    final StringBuilder results = new StringBuilder();

    // print header
    Metrics currentMetrics;
    String className;
    final Set<String> allClasses = new HashSet<String>();

    // iterate through all existing instances
    for (int i = 0; i < ALL_METRICS.size(); i++) {
        results.append(InOutUtils.LINE_SEPARATOR + InOutUtils.LINE_SEPARATOR);
        currentMetrics = ALL_METRICS.get(i);

        // printCommon is called once for each subclass
        className = currentMetrics.getClass().getName();

        if (!allClasses.contains(className)) {
            allClasses.add(className);
            results.append(className + " Class:" + InOutUtils.LINE_SEPARATOR);
            results.append(currentMetrics.printCommon());
            results.append(InOutUtils.LINE_SEPARATOR);
            results.append(InOutUtils.LINE_SEPARATOR);
        } // if

        // printMetrics is called for each instance
        results.append(currentMetrics.printMetrics());
    } // for

    return results;
} // method

/**
 * A standard <code>get*()</code>. <p>
 * Created (17.01.2002 14:38:36) by Christian Siefkes.<p>
 *
 * @return the value of the attribute
 */
protected static Configuration getConfig() {
    return config;
} // method

/**
 * Prepares the data common to all instances of a subclass
 * (i.e, static data) for printing, calculating appropriate statistics. <p>
 * Created (16.01.2002 12:14:37) by Christian Siefkes.<p>
 *
 * @return a multi-line String containing the data and statistics to print
 */
abstract String printCommon();

/**
 * Prepares the data stored in this instance for printing,
 * calculating appropriate statistics. <p>
 * Created (18.12.2001 11:02:06) by Christian Siefkes.<p>
 *
 * @return a multi-line String containing the data and statistics to print
 */
abstract String printMetrics();

/**
 * A standard <code>set*()</code>. <p>
 * Created (17.01.2002 14:38:54) by Christian Siefkes.<p>
 *
 * @param newConfig the new value of the attribute
 */
private static void setConfig(final Configuration newConfig) {
    config = newConfig;
} // method

/**
 * Removes all registered instances and resets all static fields
 * (so their results will not be collected and published any more). <p>
 *
 * Created (05.02.2002 12:28:56) by Christian Siefkes.<p>
 */
public static void reset() {
    synchronized (ALL_METRICS) {
        ALL_METRICS.clear();
    } // synchronized
} // method
} // class
