/*
 * Copyright (C) 2003-2005 Christian Siefkes <christian@siefkes.net>
 * Parts of this code are Copyright (C) 2001-2002 by MUNDWERK AG.
 * Re-released under GNU GPL with permission of MUNDWERK AG.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.predict;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import net.siefkes.nlstego.NLStego;
import static net.siefkes.nlstego.NLStego.LOG;
import net.siefkes.nlstego.util.ExtendedLRUMap;
import net.siefkes.nlstego.util.Util;

import org.apache.commons.configuration.Configuration;

/**
 * Implementation of the FetchStats class using a
 * Prediction-by-Partial-Match strategy.
 * This class collects statistics about the item access history
 * that can be used for prediction. <p>
 * Created (04.10.2001 11:16:19).<p>
 *
 * @author Christian Siefkes
 * @version $Revision: 1.14 $, $Date: 2005/07/28 15:03:09 $, $Author: siefkes $
 */
public final class PPMFetchStats extends FetchStats {

    /**
     * A constant with the name of this class. <p>
     * Created (04.10.2001 11:47:28) by Christian Siefkes.<p>
     */
    private static final String NAME = "PPMFetchStats";

    /**
     * Array of caches that store the the data used for predicting.
     * The n-th field of this array uses the n+1 previous items (ArrayList
     * of Integer) as key and a FetchStatsEntry as value. The length
     * of this array is <code>lengthOfHistory</code>. <p>
     * Created (26.09.2001 17:48:58) by Christian Siefkes.<p>
     */
    private final ExtendedLRUMap[] entryCaches;

    /**
     * Stateless model containing predictions sorted by global likelihood;
     * <code>null</code> if no stateless model should be used.
     */
    private final PPMFetchStatsEntry statelessModel;

    /**
     * The length of the session history, i.e. the number of previous
     * item considered when issuing predictions. <p>
     * Created (26.09.2001 17:54:23) by Christian Siefkes.<p>
     */
    private final int lengthOfHistory;

    /**
     * The size of the inner cache (inside each FetchStatsEntry). <p>
     * Created (01.10.2001 11:15:41) by Christian Siefkes.<p>
     */
    private final int innerCacheSize;

    /**
     * The size of the largest outer cache (caches the FetchStatsEntry's
     * using <code>lengthOfHistory</code> documents as path). <p>
     * Created (01.10.2001 11:15:45) by Christian Siefkes.<p>
     */
    private final int largestOuterCacheSize;

    /**
     * The size of each smaller outer cache is decreased by this factor (the
     * size of cache for <code>lengthOfHistory-1</code> items is
     * <code>largestOuterCacheSize/decFactor</code>; for
     * <code>lengthOfHistory-2</code> documents
      * <code>largestOuterCacheSize/(decFactor*decFactor) etc.</code>). <p>
     * Created (01.10.2001 11:15:49) by Christian Siefkes.<p>
     */
    private final double decFactor;

    /**
     * The number of past histories of different lengths blended (mixed) in
      * <code>getMostProbable()</code>.<p>
     * Created (01.10.2001 11:42:22) by Christian Siefkes. <p>
     */
    private final int blendingNum;

    /**
     * The weight of each shorter past history considered by
      * <code>getMostProbable()</code> is decreased by this factor
      * -- the default, 2.0, shouldn't be changed to get correct results. <p>
     * Created (01.10.2001 11:42:26) by Christian Siefkes.<p>
     */
    private static final double BLENDING_FACTOR = 2.0;

    /**
     * How many accesses an item must have had to be considered worth
     * of predicting. <p>
     * Created (01.10.2001 11:42:31) by Christian Siefkes.<p>
     */
    private final int minAccesses;

    /**
     * Stores the session histories of all sessions currently in progress.
     * Uses the session ID as key (Integer) and the last
     * <code>lengthOfHistory</code> items as value (ArrayList of Integer). <p>
     * Created (01.10.2001 17:09:50) by Christian Siefkes.<p>
     */
    private final Hashtable<Integer, ArrayList<Integer>> sessionHistory =
        new Hashtable<Integer, ArrayList<Integer>>();

/**
 * Creates a new instance. <p>
 * Created (11.09.2001 16:35:54) by Christian Siefkes.<p>
 *
 * @param conf the <code>Configuration</code> to use
 * @param filter optional filter that decides which predictions should be stored
 * in the stateless (global) model; if <code>null</code>, all predictions are
 * stored
 */
public PPMFetchStats(final Configuration conf, final StatelessFilter filter) {
    super(conf, filter);

    // configure fields
    lengthOfHistory = conf.getInteger("predict.historylength", 4);
    innerCacheSize = conf.getInteger("predict.innercache.size", 32);
    largestOuterCacheSize = conf.getInteger(
            "predict.outercaches.largestsize", 160000);
    decFactor = conf.getDouble("predict.outercaches.decfactor", 2.5);
    minAccesses = conf.getInteger("predict.minimumaccesses", 2);
    blendingNum = conf.getInteger("predict.blending.number", 2);

    // initializing the SizedCaches
    entryCaches = new ExtendedLRUMap[lengthOfHistory];
    int outerCacheSize = largestOuterCacheSize;

    for (int counter = entryCaches.length - 1; counter >= 0; counter--) {
        entryCaches[counter] = new ExtendedLRUMap(outerCacheSize);
        outerCacheSize /= decFactor;
    } // for

    // init stateless entry, if configured (same size as largest outer cache)
    if (conf.getBoolean("predict.stateless-model", true)) {
        statelessModel = new PPMFetchStatsEntry(largestOuterCacheSize);
    } else {
        statelessModel = null;
    }
} // method

/**
 * Adds a request of an item.<p>
 * Created (27.09.2001 12:16:43) by Christian Siefkes.<p>
 *
 * @param key unique ID of the item currently requested
 * fetching the last document in this session
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 * @param train <code>true</code> iff the model should be trained, i.e.
 * this request should be incorporated into the prediction model; otherwise
 * only the internal state will be changed so the next call to
 * {@link #predict(int, int)} will find suitable predictions (this happens in
 * any case)
 */
public void add(final String key, final int sessionID, final boolean train) {
        final Integer sessionPort = Integer.valueOf(sessionID);
    final Integer mappedKey = mapKey(key);

    // retrieving session data
    ArrayList<Integer> previousMappedKeys = sessionHistory.get(sessionPort);

    if (previousMappedKeys == null) {
        // this is the start of a session
        previousMappedKeys = new ArrayList<Integer>();

        if (lengthOfHistory > 1) {
            // otherwise only the current document can be stored
            previousMappedKeys.add(START_OF_SESSION);
        } // if

        // adding the session to the Hashtable
        sessionHistory.put(sessionPort, previousMappedKeys);
    } // if

    if (train) {
        // delegating to private add which does the actual training
        add(previousMappedKeys, mappedKey, key);
    }

    // adding this item at the end of the session history
    previousMappedKeys.add(mappedKey);

    // remove the first element of the session history, if necessary
    if (previousMappedKeys.size() > lengthOfHistory) {
        previousMappedKeys.remove(0);
    } // if
} // method

/**
 * Adds a request of an item.<p>
 * Created (27.09.2001 12:16:43) by Christian Siefkes.<p>
 *
 * @param previousMappedKeys a list of Integer's containing
 * the unique IDs of the items previously
 * requested in this session. Use the special
 * key <code>START_OF_SESSION</code> as its 0th element to mark the start
 * of a session. The length of this array must be less or equal than
 * <code>getLengthOfHistory()</code> if the 0th element is
 * <code>START_OF_SESSION</code>. Otherwise, it must be equal to
 * <code>getLengthOfHistory()</code>.
 * @param currentMappedKey the mapped key of the item currently requested,
 * or <code>END_OF_SESSION</code> when this is the end of a session
 * @param currentKey the key of the current item; <code>null</code> if it is
 * <code>START_OF_SESSION</code> or <code>END_OF_SESSION</code>
 */
private void add(final ArrayList previousMappedKeys,
        final Integer currentMappedKey, final String currentKey) {
    final ArrayList<Integer> keyList = new ArrayList<Integer>();
    ArrayList storedKeys;
    FetchStatsEntry entry;
    final int size = previousMappedKeys.size();

    // update entryCaches using the last previousMappedKeys as keys
    for (int counter = 0; counter < size; counter++) {
        // adding previousMappedKeys.get(size - (counter+1))
        // at the start of the ArrayList
        keyList.add(0, (Integer) previousMappedKeys.get(size - (counter + 1)));

        // create a clone that won't be modified later on
        storedKeys = (ArrayList) keyList.clone();

        // retrieve the entry in entryCaches[counter]
        entry = (FetchStatsEntry) entryCaches[counter].get(storedKeys);

        // create and add a new entry, if necessary
        if (entry == null) {
            entry = new PPMFetchStatsEntry(innerCacheSize);
            entryCaches[counter].put(storedKeys, entry);
/*            debugOrInfo(NAME, methodName, "Created entry in entryCaches["
                + counter + "] for keys=" + storedKeys); */
        } // if

        // add the new data to the entry
        entry.add(currentMappedKey);
/*        debugOrInfo(NAME, methodName, "Added key " + currentMappedKey
            + " in entryCaches[" + counter + "] to entry " + entry); */
    } // for

    // update stateless model, if used
    if (statelessModel != null) {
        // query stateless filter, if used (for non-null keys)
        if ((getStatelessFilter() == null) || (currentKey == null)
                || (getStatelessFilter().acceptForStateless(currentKey))) {
            statelessModel.add(currentMappedKey);
        }
    }
} // method

/**
 * A standard <code>get*()</code>. <p>
 * Created (01.10.2001 11:35:12) by Christian Siefkes.<p>
 *
 * @return the value of the attribute
 */
public int getBlendingNum() {
    return blendingNum;
} // method

/**
 * A standard <code>get*()</code>. <p>
 * Created (01.10.2001 11:16:32) by Christian Siefkes.<p>
 *
 * @return the value of the attribute
 */
public double getDecFactor() {
    return decFactor;
} // method

/**
 * A standard <code>get*()</code>. <p>
 * Created (01.10.2001 11:16:23) by Christian Siefkes.<p>
 *
 * @return the value of the attribute
 */
public int getInnerCacheSize() {
    return innerCacheSize;
} // method

/**
 * A standard <code>get*()</code>. <p>
 * Created (01.10.2001 11:16:13) by Christian Siefkes.<p>
 *
 * @return the value of the attribute
 */
public int getLargestOuterCacheSize() {
    return largestOuterCacheSize;
} // method

/**
 * A standard <code>get*()</code>. <p>
 * Created (26.09.2001 17:52:52) by Christian Siefkes.<p>
 *
 * @return the value of the attribute
 */
public int getLengthOfHistory() {
    return lengthOfHistory;
} // method

/**
 * A standard <code>get*()</code>. <p>
 * Created (01.10.2001 11:34:51) by Christian Siefkes.<p>
 *
 * @return the value of the attribute
 */
public int getMinAccesses() {
    return minAccesses;
} // method

/**
 * Predicts the items that will probably be requested next. <p>
 * Created (27.09.2001 12:16:43) by Christian Siefkes.<p>
 *
 * @param number the number of results to return, >0
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 * @return a SortedMap containing mapped keys as keys and FetchStatsResult's as
 * values. The length of the map is max (<code>number</code>, all matching
 * documents that had at least <code>minAccesses</code>)
 */
public SortedMap predict(final int number, final int sessionID) {
    final SortedMap result;
    final Integer sessionPort = Integer.valueOf(sessionID);

    // retrieving session data
    ArrayList<Integer> previousMappedKeys = sessionHistory.get(sessionPort);

    if (previousMappedKeys != null) {
        // delegating to private predict
        result = predict(previousMappedKeys, number, sessionID);
    } else {
        // no previous accesses: nothing to do!
        result = new TreeMap();
/*        debugOrInfo(NAME, "predict", sessionID,
            "no previous accesses: returning empty map"); */
    } // else

    return result;
} // method

/**
 * Predicts the items that will be probably requested next. <p>
 * Created (27.09.2001 12:16:43) by Christian Siefkes.<p>
 *
 * @param previousMappedKeys an ArrayList of Integer's containing
 * the mapped keys of the items previously
 * requested in this session. Use the special
 * key <code>START_OF_SESSION</code> as its 0th element to mark the start
 * of a session. The length of this array must be less or equal than
 * <code>getLengthOfHistory()</code> if the 0th element is
 * <code>START_OF_SESSION</code>. Otherwise, it must be equal to
 * <code>getLengthOfHistory()</code>.
 * @param number the number of results to return, >0
 * @param sessionID the session ID of this request
 * @return a SortedMap containing mapped keys as keys and FetchStatsResult's as
 * values. The length of the map is max(<code>number</code>, all matching
 * documents that had at least <code>minAccesses</code>)
 */
private SortedMap<Integer, FetchStatsResult> predict(
        final ArrayList<Integer> previousMappedKeys, final int number,
        final int sessionID) {
    final ArrayList<SortedMap<Integer, FetchStatsResult>> resultList =
        new ArrayList<SortedMap<Integer, FetchStatsResult>>();
    SortedMap<Integer, FetchStatsResult> result;
    FetchStatsEntry entry;
    ArrayList storedKeys;

    // create a clone of this list to preserve the original parameter
    final ArrayList keyList = (ArrayList) previousMappedKeys.clone();

    // look up the results to blend
    while ((resultList.size() < blendingNum) && (keyList.size() > 0)) {
        // create a clone that won't be modified later on
        storedKeys = (ArrayList) keyList.clone();

        // retrieve the entry in entryCaches[keyList.size()-1]
        entry = (FetchStatsEntry)
                entryCaches[keyList.size() - 1].transparentGet(storedKeys);
/*        debugOrInfo(NAME, methodName, sessionID, "keys=" + storedKeys
            + ", entryCaches[" + (keyList.size() - 1) + "] entry=" + entry); */

        // call predict on this entry, if it's not null
        if (entry != null) {
            // sessionID must be unique among all parallel sessions
            result = entry.predict(number, minAccesses, sessionID);
/*            debugOrInfo(NAME, methodName, sessionID,
                "prediction of this entry=" + result); */

            // add this to the resultList, if it contains any elements
            if (result.size() > 0) {
                resultList.add(result);
            } // if
        } // if

        // remove the first element of this ArrayList
        keyList.remove(0);
    } // while

/*    debugOrInfo(NAME, methodName, sessionID,
        "number of results=" + resultList.size()); */

    if (resultList.size() > 0) {
         result = resultList.get(0);

        // we have to blend the results
        if (resultList.size() > 1) {
            // 2^(resultList.size()-1) / ((2^resultList.size()) - 1)
            double blendingWeight = Math.pow(2, resultList.size() - 1)
                / (Math.pow(2, resultList.size()) - 1);
            TreeMap nextMap;
            Set keySet;
            Iterator iterator;
            Integer nextKey;
            FetchStatsResult statsResult, nextStatsResult;
            double newValue;

            keySet = result.keySet();
            iterator = keySet.iterator();

            // iterating through the first set
            while (iterator.hasNext()) {
                nextKey = (Integer) iterator.next();
                nextStatsResult = result.get(nextKey);

                // adjusting the weight
                newValue = nextStatsResult.getValue() * blendingWeight;
                nextStatsResult.setValue(newValue);
            } // while

            for (int counter = 1; counter < resultList.size(); counter++) {
                // the weight of each additional list is reduced by 2
                // (BLENDING_FACTOR)
                blendingWeight /= BLENDING_FACTOR;
                nextMap = (TreeMap) resultList.get(counter);
                keySet = nextMap.keySet();
                iterator = keySet.iterator();

                // iterating through the set
                while (iterator.hasNext()) {
                    nextKey = (Integer) iterator.next();
                    nextStatsResult = (FetchStatsResult) nextMap.get(nextKey);

                    // adjusting the weight
                    newValue = nextStatsResult.getValue() * blendingWeight;
                    nextStatsResult.setValue(newValue);

                    if (result.containsKey(nextKey)) {
                        statsResult = result.get(nextKey);

                        // adding the adjusted values
                        newValue = statsResult.getValue()
                            + nextStatsResult.getValue();
                        statsResult.setValue(newValue);
                    } else {
                        // adding this entry
                        result.put(nextKey, nextStatsResult);
                    } // else
                } // while
            } // for

/*            debugOrInfo(NAME, methodName, sessionID,
                "blended result (incl. superfluous entries)=" + result); */

            // remove superfluous entries
            final int difference = result.size() - number;
            if (difference > 0) {
                removeSuperfluousEntries(result, number, difference);
            } // if
        } // if
    } else {
        if (statelessModel != null) {
            // use stateless model as fallback
            result = statelessModel.predict(number, minAccesses, sessionID);

            if (LOG.isDebugEnabled()) {
                final List<String> unmappedResults =
                    unmapAllResults(result.values());
                final List<String> previousKeys =
                    unmapAllKeys(previousMappedKeys);
                LOG.debug("Falling back to stateless model: " + unmappedResults
                        + "; previous keys: " + previousKeys);
            }
        } else {
            // no results were found: return an empty TreeMap
            result = new TreeMap<Integer, FetchStatsResult>();
        }
    } // else

/*     debugOrInfo(NAME, methodName, sessionID,
        "returning result: " + result); */
    return result;
} // method

/**
 * Prints statistical information about this model. This method might
 * take some time while the data is gathered.
 *
 * @return a string listing statistical details about this model
 */
public String printModelStats() {
    final StringBuilder result = new StringBuilder(toString());

    if (statelessModel != null) {
        result.append(Util.LINE_SEPARATOR);
        result.append("Model 0 (stateless): ");
        final ExtendedLRUMap statelessCache = statelessModel.getEntryCache();
        result.append("size/max.size: ").append(statelessCache.size());
        result.append("/").append(statelessCache.maxSize());
    }

    for (int counter = 0; counter < entryCaches.length; counter++) {
        result.append(Util.LINE_SEPARATOR);
        result.append("Model ").append(counter + 1).append(": ");
        printModelStats(entryCaches[counter], result);
    } // for

    return result.toString();
} // method

/**
 * Prints statistical information about a model. This method might
 * take some time while the data is gathered.
 *
 * @param model the model to describe
 * @param appender statistical details about the model will be appende to this
 * object
 */
private void printModelStats(final ExtendedLRUMap model,
        final StringBuilder appender) {
    final int size = model.size();
    appender.append("size/max.size: ").append(size);
    appender.append("/").append(model.maxSize());

    PPMFetchStatsEntry entry;
    ExtendedLRUMap entryCache;
    int fullEntries = 0;
    long items = 0;
    final Iterator valueIter = model.values().iterator();

    while (valueIter.hasNext()) {
        entry = (PPMFetchStatsEntry) valueIter.next();
        entryCache = entry.getEntryCache();
        items += entryCache.size();

        if (entryCache.isFull()) {
            fullEntries++;
        }
    }

    appender.append(", full entries: ").append(fullEntries);
    appender.append(", items: ").append(items);

    final double avgItems = (double) items / size;
    appender.append(", avg. items per entry: ").append(
            Util.formatDouble(avgItems));
}

/**
 * Removes superfluous entries of a map of prediction results.
 *
 * @param result the map of entries to clean-up
 * @param number the number of results to keep, >0
 * @param difference the number of entries to remove
 */
private void removeSuperfluousEntries(final SortedMap result,
        final int number, final int difference) {
    // create a new TreeMap ordered by values
    final TreeMap<Double, List<Integer>> invertedMap =
        new TreeMap<Double, List<Integer>>();
    Set keySet = result.keySet();
    Iterator iterator = keySet.iterator();
    Integer nextKey;
    FetchStatsResult nextStatsResult;
    Double nextValue;
    List<Integer> currentKeyList;

    while (iterator.hasNext()) {
        nextKey = (Integer) iterator.next();
        nextStatsResult = (FetchStatsResult) result.get(nextKey);
        nextValue = new Double(nextStatsResult.getValue());
        currentKeyList = invertedMap.get(nextValue);

        if (currentKeyList == null) {
            currentKeyList = new ArrayList<Integer>();
        } // if

        // using the values as keys and lists of keys as values ;-)
        currentKeyList.add(nextKey);
        invertedMap.put(nextValue, currentKeyList);
    } // while

    final Set keyListSet = invertedMap.keySet();
    iterator = keyListSet.iterator();
    List nextKeyList = null;
    boolean goOn = true;
    int itemsRemoved = 0;

    // removing least valueable entries
    while (goOn) {
        while (goOn && ((nextKeyList == null)
                || (nextKeyList.isEmpty()))) {
            if (iterator.hasNext()) {
                nextValue = (Double) iterator.next();
                nextKeyList = invertedMap.get(nextValue);
            } else {
                NLStego.LOG.error("Problem while blending results: "
                    + "Could not remove the " + difference
                    + " least valuable item, because there only "
                    + "were " + itemsRemoved + " items");
                goOn = false;
            } // else
        } // if

        if (goOn) {
            // there must be at least 1 element because we checked
            // above
            nextKey = (Integer) nextKeyList.remove(0);
            result.remove(nextKey);
            itemsRemoved++;

            if (itemsRemoved >= difference) {
                // done
                goOn = false;
            } // if
        } // if
    } // while

    // check that the size is new correct
    if ((result.size() - number) != 0) {
        NLStego.LOG.error("Removing least valueable entries didn't "
            + "work correctly: " + result.size()
            + " items were retained instead of the requested "
            + number);
    } /* else {
        Log.TIES.info(
            "Successfully removed least valueable entries");
    } */
}

/**
 * Marks a session as ended. <p>
 * Created (27.09.2001 12:16:43) by Christian Siefkes.<p>
 *
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 */
public void sessionEnded(final int sessionID) {
    final Integer sessionPort = Integer.valueOf(sessionID);

    // retrieve session data
    ArrayList previousMappedKeys = sessionHistory.get(sessionPort);

    if (previousMappedKeys != null) {
        // delegating to private add
        add(previousMappedKeys, END_OF_SESSION, null);

        // remove this session from the history Hashtable
        sessionHistory.remove(sessionPort);
    } // if
} // method

/**
 * Returns a String that represents the value of this object. <p>
 * Created (04.10.2001 17:42:12) by Christian Siefkes.<p>
 *
 * @return a string representation of the receiver
 */
public String toString() {
    return "[" + NAME + ": length of history=" + getLengthOfHistory()
        + ", inner cache size=" + getInnerCacheSize()
        + ", largest outer cache size=" + getLargestOuterCacheSize()
        + ", decrease factor=" + getDecFactor()
        + ", blending number=" + getBlendingNum()
        + ", minimum accesses=" + getMinAccesses() + "]";
} // method
} // class
