/*
 * Copyright (C) 2003-2005 Christian Siefkes <christian@siefkes.net>
 * Parts of this code are Copyright (C) 2001-2002 by MUNDWERK AG.
 * Re-released under GNU GPL with permission of MUNDWERK AG.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.predict;

import java.util.Iterator;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.TreeSet;

import static net.siefkes.nlstego.NLStego.LOG;
import net.siefkes.nlstego.util.ExtendedLRUMap;

/**
 * Implementation of the FetchStatsEntry class using a
 * Prediction-by-Partial-Match strategy.
 * Contains information on the items accessed in a certain state.<p>
 * Created (04.10.2001 11:16:42).<p>
 *
 * @author Christian Siefkes
 * @version $Revision: 1.9 $, $Date: 2005/07/12 17:02:17 $, $Author: siefkes $
 */
final class PPMFetchStatsEntry extends FetchStatsEntry {

    /**
     * A constant with the name of this class.<p>
     * Created (04.10.2001 11:47:20) by Christian Siefkes.<p>
     */
    private static final String NAME = "PPMFetchStatsEntry";

    /**
     * Used to guard synchronization.<p>
     * Created (17.09.2001 16:22:17) by Christian Siefkes.<p>
     */
    private final Object guard = new Object();

    /**
     * Stores the number of access of an item. Uses the item ID
     * as key and a FetchStatsSubEntry as value.<p>
     * Created (17.09.2001 16:28:27) by Christian Siefkes.<p>
     */
    private final ExtendedLRUMap entryCache;

    /**
     * Sorts the items by the frequency they were accessed.
     * Uses the number of accesses (Integer) as key and a TreeSet of item
     * IDs as value.<p>
     * Created (17.09.2001 17:30:02) by Christian Siefkes.<p>
     */
    private final TreeMap<Integer, TreeSet<Integer>> frequencyMap =
        new TreeMap<Integer, TreeSet<Integer>>();

    /**
     * The sum of all accesses stored for this item.<p>
     * Created (18.09.2001 15:00:34) by Christian Siefkes.<p>
     */
    private int sumOfAccesses = 0;

/**
 * Creates a new instance.<p>
 * Created (17.09.2001 16:08:57) by Christian Siefkes.<p>
 *
 * @param numberOfEntries the maximum number of entries cached
 */
PPMFetchStatsEntry(final int numberOfEntries) {
    super();
    entryCache = new ExtendedLRUMap(numberOfEntries);
} // constructor

/**
 * Adds a request of an item.<p>
 * Created (17.09.2001 16:15:49) by Christian Siefkes.<p>
 *
 * @param key the unique ID of the item
 */
void add(final Integer key) {
    FetchStatsSubEntry entry = null;
    Integer frequencyKey;
    TreeSet<Integer> frequencyEntry;

    synchronized (guard) {
        entry = (FetchStatsSubEntry) entryCache.get(key);

        // do we know this ID?
        if (entry == null) {
            // Cleanup, if necessary:
            if (entryCache.isFull() && !entryCache.containsKey(key)) {
/*                Log.TIES.debug("Entry cache is full "
                    + "- cleaning up least-recently-used items"); */
                FetchStatsSubEntry deletedSubEntry;
                int deletedAccesses;

                deletedSubEntry =
                    (FetchStatsSubEntry) entryCache.transparentGet(
                            entryCache.firstKey());
                deletedAccesses = deletedSubEntry.getAccesses();

                // decreasing the sum of accesses:
                sumOfAccesses -= deletedAccesses;

                // removing deleted entry from frequencyMap:
                frequencyKey = Integer.valueOf(deletedAccesses);
                frequencyEntry = frequencyMap.get(frequencyKey);
                frequencyEntry.remove(key);
            } // if

            // creating new entry for this ID
            entry = new FetchStatsSubEntry();
            entryCache.put(key, entry);
            frequencyKey = Integer.valueOf(1);
        } else {
            // retrieving data
            int accesses = entry.getAccesses();

            // increasing number of accesses by one
            int newAccesses = accesses + 1;

            // storing updated data
            entry.setAccesses(newAccesses);

            // removing old entry from frequencyMap
            frequencyKey = Integer.valueOf(accesses);
            frequencyEntry = frequencyMap.get(frequencyKey);
            frequencyEntry.remove(key);

            frequencyKey = Integer.valueOf(newAccesses);
        } // else

        // increasing the sum of accesses by one
        sumOfAccesses++;

        // creating new entry in frequencyMap
        frequencyEntry = frequencyMap.get(frequencyKey);

        if (frequencyEntry == null) {
            // initializing and adding the TreeSet
            frequencyEntry = new TreeSet<Integer>();
            frequencyMap.put(frequencyKey, frequencyEntry);
        } // if

        frequencyEntry.add(key);

    } // synchronized
} // method

    /**
     * Returns the entry cache used by instance for inspection purposes.
     *
     * @return the value of the attribute
     */
    ExtendedLRUMap getEntryCache() {
        return entryCache;
    }

/**
 * Returns the most probable items as well as their probabilities
 * (List of FetchStatsResult's).<p>
 * Created (18.09.2001 15:47:31) by Christian Siefkes.<p>
 *
 * @param number the number of results to return, > 0
 * @param minAccesses how many accesses an item must have had to be
 * considered worth of prefetching, > 0
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 * @return a SortedMap containing mapped keys as keys and FetchStatsResult's as
 * values. The size of the map is max
 * (<code>number</code>, all items stored in this entry that had at least
 * <code>minAccesses</code>)
 */
SortedMap<Integer, FetchStatsResult> predict(final int number,
        final int minAccesses, final int sessionID) {
    final SortedMap<Integer, FetchStatsResult> result =
        new TreeMap<Integer, FetchStatsResult>();

    if ((number > 0) && (frequencyMap.size() > 0) && (sumOfAccesses > 0)) {
        // otherwise we have nothing to do

        FetchStatsSubEntry entry = null;
        Integer frequencyKey = null;
        int currentFrequency = -1;
        TreeSet frequencyEntry = null;
        double currentProbability = -1;
        Iterator iterator = null;
        FetchStatsResult statsResult = null;
        Integer key = null;
        int counter = 0;

        try {
            synchronized (guard) {
                // getting the highest frequency and starting iterating it
                frequencyKey = frequencyMap.lastKey();
                currentFrequency = frequencyKey.intValue();
                frequencyEntry = frequencyMap.get(frequencyKey);
                currentProbability = ((double) currentFrequency)
                    / sumOfAccesses;
                iterator = frequencyEntry.iterator();

                while ((counter < number)
                    && (currentFrequency >= minAccesses)) {

                    if (iterator.hasNext()) {
                        // retrieving the data needed to create this
                        // FetchStatsResult
                        key = (Integer) iterator.next();

                        // negative keys (FetchStats.START/END_OF_SESSION
                        // are ignored
                        if (key.intValue() >= 0) {
                            // transparentGet, as predictions should not modify
                            // the replacement structure
                            entry = (FetchStatsSubEntry)
                                entryCache.transparentGet(key);

                            if (entry != null) {
                                statsResult = new FetchStatsResult(
                                    key.intValue(), currentProbability,
                                    sessionID);

                                // inserting the FetchStatsResult in the result
                                // TreeMap
                                result.put(key, statsResult);
                                counter++;
                            } // if
                        } // if
                    } else {
                        // we're at the end of this TreeMap, switching to the
                        // next highest
                        do {
                            currentFrequency--;
                            frequencyKey = Integer.valueOf(currentFrequency);
                            frequencyEntry = frequencyMap.get(frequencyKey);

                            if (frequencyEntry != null) {
                                iterator = frequencyEntry.iterator();
                                currentProbability = ((double) currentFrequency)
                                    / sumOfAccesses;
                            } // if

                            // we have to repeat this, if frequencyEntry == null
                            // but we're done if currentFrequency < minAccesses
                        } while ((frequencyEntry == null)
                            && (currentFrequency >= minAccesses));
                    } // if
                } // while

            } // synchronized

        } catch (RuntimeException rte) {
            // for debugging purposes: logging error details before
            // rethrowing exception
            LOG.error("Unexpected exception occured during method call; "
                + "state of this object=" + toString()
                + "; method parameters: number=" + number
                + ", minAccesses=" + minAccesses + ", sessionID=" + sessionID
                + "; local variables: entry=" + entry
                + ", frequencyKey=" + frequencyKey
                + ", currentFrequency=" + currentFrequency
                + ", frequencyEntry=" + frequencyEntry
                + ", currentProbability=" + currentProbability
                + ", iterator=" + iterator
                + ", statsResult=" + statsResult
                + ", key=" + key
                + ", counter=" + counter,
                rte);
            throw rte;
        } // catch

    } // if

    return result;
} // method

/**
 * Returns a String that represents the value of this object.<p>
 * Created (14.11.2001 10:59:52) by Christian Siefkes.<p>
 *
 * @return a string representation of the receiver
 */
public String toString() {
    return "[" + NAME + ": entry cache=" + entryCache
        + ", frequency map=" + frequencyMap
        + ", sum of accesses=" + sumOfAccesses + "]";
} // method
} // class
