/*
 * Copyright (C) 2003-2005 Christian Siefkes <christian@siefkes.net>
 * Parts of this code are Copyright (C) 2001-2002 by MUNDWERK AG.
 * Re-released under GNU GPL with permission of MUNDWERK AG.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.predict;

import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import net.siefkes.nlstego.util.Util;

import org.apache.commons.configuration.Configuration;

/**
 * Manages the predictions of items, combining different prediction models.
 * <p>
 * Created (26.11.2001 09:32:34).<p>
 *
 * @author Christian Siefkes
 * @version $Revision: 1.9 $, $Date: 2005/07/12 17:02:17 $, $Author: siefkes $
 */
public final class PredictionManager {

    /**
     * A constant with the name of this class. <p>
     * Created (26.11.2001 10:39:38) by Christian Siefkes.<p>
     */
    private static final String NAME = "PredictionManager";

    /**
     * Used to manage the IDs of all instances--contains the next ID
     * that isn't yet used. The constructor of a new instance stores the
     * current value as instance ID and then increments this field.
     */
    private static int nextAvailableID = 0;

    /**
     * Used to synchronize the {@link #nextAvailableID} field.
     */
    private static final Object ID_GUARD = new Object();

    /**
     * The unique ID of this instance.
     */
    private final int id;

    /**
     * Items whose probability to be requested is smaller than this value
     * won't be predicted.
     * A probability, i.e., a value between 0 and 1. <p>
     * Created (04.10.2001 16:58:05) by Christian Siefkes.<p>
     */
    private final double minProbability;

    /**
     * The maximum number of predictions made by each model.<p>
     * Created (04.10.2001 16:57:55) by Christian Siefkes.<p>
     */
    private final int numberOfPredictions;

    /**
     * A list of prediction models ({@link FetchStats}).
     */
    private final List<FetchStats> predictionModelList =
        new LinkedList<FetchStats>();

    /**
     * Contains model IDs (Integer) as keys and HashMaps of mappedKey
     * Integers/FetchStatsResults as values.
     * Used to retrieve the results and to delete model requests. <p>
     * Created (25.10.2001 16:40:36) by Christian Siefkes.<p>
     */
    private final Map<Integer, Map<Integer, FetchStatsResult>>
    predictionsByModel = new HashMap<Integer, Map<Integer, FetchStatsResult>>();

    /**
     * Contains mapped keys (Integer) as keys and HashSets of model IDs
     * (Integers) as values.
     * Used to combine the values of keys that are predicted several times.
     * <p>
     * Created (25.10.2001 16:40:36) by Christian Siefkes.<p>
     */
    private final Map<Integer, Set<Integer>> predictionsByKey =
        new HashMap<Integer, Set<Integer>>();

    /**
     * Contains inverted combined values (Double) as keys and HashSets of
     * mapped keys (Integer) as values.
     * Combines the values of keys that are predicted by several models.
     * The most valuable keys come first (their inverted value is lowest).
     * <p>
     * Created (25.10.2001 16:40:36) by Christian Siefkes.<p>
     */
    private final SortedMap<Double, Set<Integer>> predictionsByInvertedValue =
        new TreeMap<Double, Set<Integer>>();

    /**
     * Contains mapped keys (Integer) as keys and inverted combined values
     * (Double) as values.
     * Used to retrieve the inverted combined values. <p>
     * Created (25.10.2001 16:40:36) by Christian Siefkes.<p>
     */
    private final Map<Integer, Double> invertedValues =
        new HashMap<Integer, Double>();

    /**
     * Measure the prediction quality?
     * Depends on the "metrics" parameter. <p>
     * Created (02.01.2002 17:20:46) by Christian Siefkes.<p>
     */
    private boolean useMetrics;

//    /**
//     * Used to measure the prediction quality. <p>
//     * Created (02.01.2002 17:12:27) by Christian Siefkes.<p>
//     */
//    private PredictionMetrics metrics;

    /**
     * Whether the predictions are currently up-to-date or out-of-date.
     */
    private boolean uptodatePredictions = true;

/**
 * Creates a new instance. <p>
 * Created (26.11.2001 09:34:51) by Christian Siefkes.<p>
 *
 * @param conf the <code>Configuration</code> to use
 */
public PredictionManager(final Configuration conf) {
    super();

    // determine ID, afterwards incrementing static ID counter
    synchronized (ID_GUARD) {
        id = nextAvailableID++;
    }

    // configure fields
    minProbability = conf.getDouble("predict.probability.minimum", 0.01);
    numberOfPredictions = conf.getInteger("predict.predictions.number", 12);

    // measure the prediction quality?
    setUseMetrics(conf.getBoolean("predict.metrics", true));
    //initMetrics(conf);
} // constructor

/**
 * Adds a key to the predictions of a model -- all fields are handled. <p>
 * Created (27.11.2001 09:57:40) by Christian Siefkes.<p>
 *
 * @param modelID the unique ID of the model
 * @param mappedKey the representation of the key to remove
 * @param fsResult the FetchStatsResult of the key to remove
 */
private void addKey(final Integer modelID, final Integer mappedKey,
    final FetchStatsResult fsResult) {

    Set<Integer> keyModels = predictionsByKey.get(mappedKey);

    if (keyModels == null) {
        // this key wasn't predicted by other models
        keyModels = new HashSet<Integer>();
        predictionsByKey.put(mappedKey, keyModels);
    } // if

    keyModels.add(modelID);

    Double oldInvertedValue = invertedValues.get(mappedKey);
    Double newInvertedValue;
    Set<Integer> keysWithThisValue;

    if (oldInvertedValue == null) {
        newInvertedValue = new Double(-fsResult.getValue());
    } else {
        newInvertedValue = new Double(oldInvertedValue.doubleValue()
            - fsResult.getValue());

        keysWithThisValue = predictionsByInvertedValue.get(oldInvertedValue);

        if (keysWithThisValue.size() == 1) {
            // no other keys with this value
            predictionsByInvertedValue.remove(oldInvertedValue);
        } else {
            keysWithThisValue.remove(mappedKey);
        } // else
    } // else

    invertedValues.put(mappedKey, newInvertedValue);

    keysWithThisValue = predictionsByInvertedValue.get(newInvertedValue);

    if (keysWithThisValue == null) {
        // create and add this HashSet
        keysWithThisValue = new HashSet<Integer>();
        predictionsByInvertedValue.put(newInvertedValue, keysWithThisValue);
    } // if

    keysWithThisValue.add(mappedKey);

    Map<Integer, FetchStatsResult> predictionsOfThisModel =
        predictionsByModel.get(modelID);

    if (predictionsOfThisModel == null) {
        // first prediction of this model: create and add HashMap
        predictionsOfThisModel = new HashMap<Integer, FetchStatsResult>();
        predictionsByModel.put(modelID, predictionsOfThisModel);
    } // if

    predictionsOfThisModel.put(mappedKey, fsResult);
} // method

/**
 * Updates the predictions of a model.  For performance reasons,
 * this method is not synchronized so you are not allowed to register
 * or unregister prediction models during a call to this method. <p>
 * Created (26.11.2001 09:40:18) by Christian Siefkes.<p>
 *
 * @throws IllegalStateException if this method was called but there are no
 * registered prediction models
 * @throws ConcurrentModificationException might be thrown if prediction
 * models are (un)registered during a call to this method
 *
 * @see #registerPredictionModel(FetchStats)
 * @see #unregisterPredictionModel(FetchStats)
 */
private void updatePredictions()
        throws IllegalStateException, ConcurrentModificationException {
    if (predictionModelList.isEmpty()) {
            throw new IllegalStateException("No prediction models registered");
    }

    // update predictions, combinding all prediction models
    final Iterator modelIterator = predictionModelList.iterator();
    FetchStats predictionModel;
    SortedMap predictions;
    Set keySet;
    Integer nextKey;
    FetchStatsResult nextResult;
    Iterator predictionIterator;
    int modelID;
    Integer modelIDInteger;

    while (modelIterator.hasNext()) {
        predictionModel = (FetchStats) modelIterator.next();
        modelID = predictionModel.getID();
        modelIDInteger = predictionModel.getIDInteger();

        // delete the old predictions
        clearModelPredictions(modelIDInteger);

        predictions = predictionModel.predict(getNumberOfPredictions(),
            modelID);
        keySet = predictions.keySet();
        predictionIterator = keySet.iterator();

        // add new predictions
        while (predictionIterator.hasNext()) {
            nextKey = (Integer) predictionIterator.next();
            nextResult = (FetchStatsResult) predictions.get(nextKey);
            addKey(modelIDInteger, nextKey, nextResult);
        } // while
    }

    /* update the metrics
    if (isUseMetrics()) {
        getMetrics().addPredictions(modelID, keySet);
    } // if
    */

    // predictions are up-to-date
    uptodatePredictions = true;
} // method


/**
 * Delegates to {@link #choosePredictions(int)}, setting the number of
 * predictions to return to {@link #getNumberOfPredictions()}.
 *
 * @return a list of {@link PredictedToken}s with the most valuable
 * predictions; will have between 0 (if no more predictions with the minimum
 * probability exist) and {@link #getNumberOfPredictions()} elements; sorted
 * according to their probability
 * @throws IllegalStateException if this method was called but there are no
 * registered prediction models
 * @throws ConcurrentModificationException might be thrown if prediction
 * models are (un)registered during a call to this method
 */
public List<PredictedToken> choosePredictions()
throws IllegalStateException, ConcurrentModificationException {
    return choosePredictions(getNumberOfPredictions());
}

/**
 * Selects the most valuable predictions. For performance reasons,
 * this method is not synchronized so you are not allowed to
 * register or unregister prediction models during a call to this method. <p>
 * Created (26.11.2001 09:53:36) by Christian Siefkes.<p>
 *
 * @param number the number of predictions to return
 * @return a list of {@link PredictedToken}s with the most valuable
 * predictions; will have between 0 (if no more predictions with the minimum
 * probability exist) and <code>number</code> elements; sorted according to
 * their probability
 * @throws IllegalStateException if this method was called but there are no
 * registered prediction models
 * @throws ConcurrentModificationException might be thrown if prediction
 * models are (un)registered during a call to this method
 *
 * @see #registerPredictionModel(FetchStats)
 * @see #unregisterPredictionModel(FetchStats)
 */
public List<PredictedToken> choosePredictions(final int number)
        throws IllegalStateException, ConcurrentModificationException {
    List<PredictedToken> result = new LinkedList<PredictedToken>();
    Iterator innerIterator;
    /*
    Iterator modelIterator;
    Integer currentModel;
    Map currentMap;
    FetchStatsResult currentFSResult;
    Set keyIsInModels;
    */
    Double currentInvertedValue;
    double currentValue;
    Set<Integer> currentSet;
    Integer currentKey;
    PredictedToken currentToken;
    boolean finished = false;
    int itemsFound = 0;

    // update the predictions, if necessary
    if (!uptodatePredictions) {
        updatePredictions();
    }

    final Set keySet = predictionsByInvertedValue.keySet();
    final Iterator iterator = keySet.iterator();

    while (iterator.hasNext() && (itemsFound < number) && (!finished)) {
        currentInvertedValue = (Double) iterator.next();
        currentValue = -currentInvertedValue.doubleValue();

        if (currentValue <= getMinProbability()) {
            // no more valuable predictions
            finished = true;
        } else {
            currentSet = predictionsByInvertedValue.get(currentInvertedValue);
            innerIterator = currentSet.iterator();

            while (innerIterator.hasNext() && (itemsFound < number)
                    && (!finished)) {
                currentKey = (Integer) innerIterator.next();

                // add the token to the result list
                currentToken = new PredictedToken(
                    FetchStats.unmapKey(currentKey.intValue()),
                    currentValue);
                result.add(currentToken);
                itemsFound++;
            } // while
        } // else
    } // while

    return result;
} // method

/**
 * Removes the predictions of a model. Must be synchronized externally. <p>
 * Created (26.11.2001 09:50:11) by Christian Siefkes.<p>
 *
 * @param modelID the ID of the moel
 */
private void clearModelPredictions(final Integer modelID) {
    Map<Integer, FetchStatsResult> modelEntries =
        predictionsByModel.get(modelID);

    if (modelEntries != null) {
        Set keySet = modelEntries.keySet();
        Iterator iterator = keySet.iterator();
        Integer nextKey;
        FetchStatsResult nextResult;

        while (iterator.hasNext()) {
            nextKey = (Integer) iterator.next();
            nextResult = modelEntries.get(nextKey);
            removeKey(modelID, nextKey, nextResult);
        } // while

        // resetting the preditionsByModel
        predictionsByModel.remove(modelID);
    } // if
} // method

/**
 * A standard <code>get*()</code>. <p>
 * Created (04.10.2001 16:56:09) by Christian Siefkes.<p>
 *
 * @return the value of the attribute
 */
public double getMinProbability() {
    return minProbability;
} // method

    /**
     * A standard <code>get*()</code>.<p>
     * Created (04.10.2001 16:56:03) by Christian Siefkes.<p>
     *
     * @return the value of the attribute
     */
    private int getNumberOfPredictions() {
        return numberOfPredictions;
    } // method

/**
 * A standard <code>get*()</code>. <p>
 * Created (02.01.2002 17:19:45) by Christian Siefkes.<p>
 *
 * @return the value of the attribute
 */
private boolean isUseMetrics() {
    return useMetrics;
} // method

/**
 * Registers an actual request. The predictions
 * of this model are removed. After calling this method you can call
 * <code>updatePredictions</code> to add new ones. For performance reasons,
 * this method is not synchronized so you are not allowed to register
 * or unregister prediction models during a call to this method.<p>
 * Created (03.01.2002 14:02:41) by Christian Siefkes.<p>
 *
 * @param key the key of the item currently requested
 * @param train <code>true</code> iff the models should be trained, i.e.
 * this request should be incorporated into the prediction models; otherwise
 * only the internal state of the models will be changed so the next call to
 * {@link #choosePredictions(int)} will find suitable predictions (this happens
 * in any case)
 * @throws IllegalStateException if this method was called but there are no
 * registered prediction models
 * @throws ConcurrentModificationException might be thrown if prediction
 * models are (un)registered during a call to this method
 *
 * @see #registerPredictionModel(FetchStats)
 * @see #unregisterPredictionModel(FetchStats)
 */
public void addItem(final String key, final boolean train)
        throws IllegalStateException, ConcurrentModificationException {

    if (predictionModelList.isEmpty()) {
        throw new IllegalStateException("No prediction models registered");
    }

    // update all prediction models
    final Iterator modelIterator = predictionModelList.iterator();
    FetchStats predictionModel;

    while (modelIterator.hasNext()) {
        predictionModel = (FetchStats) modelIterator.next();
        predictionModel.add(key, getID(), train);
    }

    /* update the metrics
    if (isUseMetrics()) {
        getMetrics().addActualFetch(key, modelInteger);
    } // if
    */

    // predictions require updating
    uptodatePredictions = false;
} // method

/**
 * Removes a key from the predictions of a model. In addition, the key
 * must be removed from the <code>predictionsByModel</code> -- all other
 * fields are handled. <p>
 * Created (26.11.2001 09:50:11) by Christian Siefkes.<p>
 *
 * @param modelID the ID of the model
 * @param mappedKey the representation of the key to remove
 * @param fsResult the FetchStatsResult of the key to remove
 */
private void removeKey(final Integer modelID, final Integer mappedKey,
    final FetchStatsResult fsResult) {

    final Set<Integer> keyModels = predictionsByKey.get(mappedKey);
    final Double invertedValue = invertedValues.get(mappedKey);
    Set<Integer> keysWithThisValue = predictionsByInvertedValue.get(
        invertedValue);

    if (keysWithThisValue.size() == 1) {
        // no other keys with this value
        predictionsByInvertedValue.remove(invertedValue);
    } else {
        keysWithThisValue.remove(mappedKey);
    } // else

    invertedValues.remove(mappedKey);

    if (keyModels.size() == 1) {
        // the key was only predicted by this model
        predictionsByKey.remove(mappedKey);
    } else {
        keyModels.remove(modelID);
        Double newInvertedValue = new Double(invertedValue.doubleValue()
            + fsResult.getValue());
        invertedValues.put(mappedKey, newInvertedValue);
        keysWithThisValue = predictionsByInvertedValue.get(newInvertedValue);

        if (keysWithThisValue == null) {
            // create and add this HashSet
            keysWithThisValue = new HashSet<Integer>();
            predictionsByInvertedValue.put(newInvertedValue, keysWithThisValue);
        } // if

        keysWithThisValue.add(mappedKey);
    } // else
} // method

/**
 * A standard <code>set*()</code>. <p>
 * Created (02.01.2002 17:21:56) by Christian Siefkes.<p>
 *
 * @param newUseMetrics the new value of the attribute
 */
private void setUseMetrics(final boolean newUseMetrics) {
    useMetrics = newUseMetrics;
} // method

/**
 * Returns a String that represents the value of this object. <p>
 * Created (04.10.2001 17:42:12) by Christian Siefkes.<p>
 *
 * @return a string representation of the receiver
 */
public String toString() {
    return "[" + NAME + ": minimum probability=" + getMinProbability()
        + ", number of predictions=" + getNumberOfPredictions()
        + ", number of registered prediction models="
        + predictionModelList.size()
        + ", use metrics=" + isUseMetrics() + "]";
} // method

    /**
     * Returns an iterator over the registered prediction models.
     * All elements in the iterator are guaranteed to be {@link FetchStats}.
     *
     * @return an iterator over the models
     */
    public Iterator<FetchStats> getPredictionModelList() {
        return predictionModelList.iterator();
    }

    /**
     * Prints statistical information about the used models. This method might
     * take some time while the data is gathered.
     *
     * @return a string listing statistical details about the used models
     */
    public String printModelStats() {
        final StringBuilder result = new StringBuilder();
        final Iterator<FetchStats> modelIter = getPredictionModelList();

        while (modelIter.hasNext()) {
            result.append(modelIter.next().printModelStats());

            if (modelIter.hasNext()) {
                result.append(Util.LINE_SEPARATOR);
            }
        }
        return result.toString();
    }

    /**
     * Registers a prediction model with this prediction manager.
     *
     * @param model the model to register
     */
    public void registerPredictionModel(final FetchStats model) {
        synchronized (predictionModelList) {
            predictionModelList.add(model);
        }
    }

    /**
     * Unregisters a prediction model from this prediction manager.
     *
     * @param model the model to unregister
     * @return whether the model was successfully unregistered (i.e.
     * whether it has been registered before)
     */
    public boolean unregisterPredictionModel(final FetchStats model) {
        synchronized (predictionModelList) {
            return predictionModelList.remove(model);
        }
    }

    /**
     * Returns the unique ID of this instance.
     *
     * @return the ID
     */
    public int getID() {
        return id;
    }
} // class
