/*
 * Copyright (C) 2003 Christian Siefkes <christian@siefkes.net>
 * Parts of this code are Copyright (C) 2001-2002 by MUNDWERK AG.
 * Re-released under GNU GPL with permission of MUNDWERK AG.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.predict;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import net.siefkes.nlstego.util.InOutUtils;
import net.siefkes.nlstego.util.Util;

import org.apache.commons.configuration.Configuration;

/**
 * This class measures and logs the prediction quality. <p>
 * Created (20.12.2001 11:20:01).<p>
 *
 * @author Christian Siefkes
 * @version $Revision: 1.8 $, $Date: 2005/07/12 17:02:17 $, $Author: siefkes $
 */
final class PredictionMetrics extends Metrics {

    /**
     * A constant with the name of this class.<p>
     * Created (02.01.2002 16:15:50) by Christian Siefkes.<p>
     */
    private static final String NAME = "PrefetchMetrics";

    /**
     * A constant value used to retrieve predictions in the used arrays. <p>
     * Created (07.01.2002 11:40:10) by Christian Siefkes.<p>
     */
    private static final int PREDICT = 0;

    /**
     * A constant value used to retrieve prefetches in the used arrays. <p>
     * Created (07.01.2002 11:40:10) by Christian Siefkes.<p>
     */
    private static final int PREFETCH = 1;

    /**
     * Contains the name of each array element.<p>
     * Created (07.01.2002 11:42:40) by Christian Siefkes.<p>
     */
    private static final String[] NAMES = {
        "Predictions",
        "Prefetches"
        };

    /**
     * Contains the number of all predictions/prefetches. <p>
     * Created (07.01.2002 11:45:12) by Christian Siefkes.<p>
     */
    private final long[] allPre = new long[NAMES.length];

    /**
     * Contains the number of accurate predictions/prefetches. <p>
     * Created (07.01.2002 11:45:12) by Christian Siefkes.<p>
     */
    private final long[] accurate = new long[NAMES.length];

    /**
     * Contains the number of all prefetchable actual fetches. <p>
     * Created (07.01.2002 11:47:14) by Christian Siefkes.<p>
     */
    private long allActual = 0;

    /**
     * Contains sessionIDs (Integer) as keys and HashSets of key
     * Strings as values. <p>
     * Created (07.01.2002 11:43:21) by Christian Siefkes.<p>
     */
    private final Map<Integer, Set<String>> predictionsBySession =
        new HashMap<Integer, Set<String>>();

    /**
     * Contains sessionIDs (Integer) as keys and HashSets of key
     * Strings as values. <p>
     * Created (03.01.2002 16:01:23) by Christian Siefkes.<p>
     */
    private final Map<Integer, Set<String>> prefetchesBySession =
        new HashMap<Integer, Set<String>>();

/**
 * Creates a new instance. <p>
 * Created (13.12.2001 10:45:58) by Christian Siefkes.<p>
 *
 * @param conf the <code>Configuration</code> to use
 */
PredictionMetrics(final Configuration conf) {
    super(conf);
} // method

/**
 * Adds an actual (non-prefetch) fetch to the metrics. <p>
 * Created (07.01.2002 11:14:35) by Christian Siefkes.<p>
 *
 * @param key the key of the item currently requested
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 */
synchronized void addActualFetch(final String key, final Integer sessionID) {
    // update allActual
    allActual++;

    // update accurate[PREFETCH]
    Set<String> currentSet = predictionsBySession.get(sessionID);

    if ((currentSet != null) && (currentSet.contains(key))) {
        accurate[PREDICT]++;
    } // if

    // update accurate[PREFETCH]
    currentSet = prefetchesBySession.get(sessionID);

    if ((currentSet != null) && (currentSet.contains(key))) {
        accurate[PREFETCH]++;
    } // if

    // delete the predictions/prefetchesBySession of this session
    clearSession(sessionID);

} // method

/**
 * Adds the new predictions of a session to the metrics,
 * removing the old ones. <p>
 * Created (07.01.2002 11:50:33) by Christian Siefkes.<p>
 *
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 * @param predictions a Set of key Strings
 */
synchronized void addPredictions(final Integer sessionID,
        final Set<String> predictions) {
    // update allPre[PREDICT]
    allPre[PREDICT] += predictions.size();

    // update predictionsBySession, replacing the old predictions if any
    predictionsBySession.put(sessionID, predictions);

} // method

/**
 * Adds a prefetch request to the statistics. <p>
 * Created (03.01.2002 16:12:44) by Christian Siefkes.<p>
 *
 * @param key the key of the item currently requested
 * @param sessionIDs a Set containing the IDs (Integer) of the sessions
 * that predicted this key
 */
synchronized void addPrefetch(final String key, final Set sessionIDs) {
    // update allPre[PREFETCH]
    allPre[PREFETCH]++;

    // update prefetchesBySession
    final Iterator iterator = sessionIDs.iterator();
    Integer currentSession;
    Set<String> currentSet;

    while (iterator.hasNext()) {
        currentSession = (Integer) iterator.next();
        currentSet = prefetchesBySession.get(currentSession);

        // create a Set for this session, if required
        if (currentSet == null) {
            currentSet = new HashSet<String>();
            prefetchesBySession.put(currentSession, currentSet);
        } // if

        // add this key
        currentSet.add(key);
    } // while
} // method

/**
 * For internal use: removes the predictions and prefetches
 * of this session. <p>
 * Created (07.01.2002 15:17:20) by Christian Siefkes.<p>
 *
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 */
private void clearSession(final Integer sessionID) {
    predictionsBySession.remove(sessionID);
    prefetchesBySession.remove(sessionID);
} // method

/**
 * Prepares the data common to all instances of a subclass
 * (i.e, static data) for printing, calculating appropriate statistics. <p>
 * Created (17.01.2002 14:33:31) by Christian Siefkes.<p>
 *
 * @return a multi-line String containing the data and statistics to print
 */
synchronized String printCommon() {
    // currently this method doesn't print anything
    return "";
} // method

/**
 * Prepares the data stored in this instance for printing,
 * calculating appropriate statistics. <p>
 * Created (02.01.2002 16:14:52) by Christian Siefkes.<p>
 *
 * @return a multi-line String containing the data and statistics to print
 */
synchronized String printMetrics() {
    final StringBuilder result = new StringBuilder();
    double accuracy, usefulness;

    result.append(NAME + ":" + InOutUtils.LINE_SEPARATOR);

    // calculate and print accuracy and usefulness
    for (int i = 0; i < NAMES.length; i++) {
        // accuracy: percentage of prefetched/predicted pages that were later
        // actually requested
        result.append("Accuracy of " + NAMES[i] + "\t");

        if (allPre[i] > 0) {
            accuracy = ((double) accurate[i]) / allPre[i];
            result.append(Util.formatPercentage(accuracy)
                    + InOutUtils.LINE_SEPARATOR);
        } else {
            result.append("No " + NAMES[i] + InOutUtils.LINE_SEPARATOR);
        } // else

        // usefulness: percentage of fetched pages that had already been
        // prefetched/predicted (accurate[PREDICT/PREFETCH] / allActual)
        result.append("Usefulness of " + NAMES[i] + "\t");

        if (allActual > 0) {
            usefulness = ((double) accurate[i]) / allActual;
            result.append(Util.formatPercentage(usefulness)
                    + InOutUtils.LINE_SEPARATOR);
        } else {
            result.append("No actual fetches" + InOutUtils.LINE_SEPARATOR);
        } // else
    } // for

    result.append(InOutUtils.LINE_SEPARATOR);

    // print header
    result.append(InOutUtils.LINE_SEPARATOR + "Raw data:"
            + InOutUtils.LINE_SEPARATOR);

    for (int i = 0; i < NAMES.length; i++) {
        // print raw data
        result.append("All " + NAMES[i] + "\t" + allPre[i]
                + InOutUtils.LINE_SEPARATOR);
        result.append("Accurate " + NAMES[i] + "\t" + accurate[i]
                + InOutUtils.LINE_SEPARATOR);
    } // for

    // print number of fetches
    result.append("Prefetchable actual fetches\t" + allActual
            + InOutUtils.LINE_SEPARATOR);

    return result.toString();
} // method

/**
 * Marks a session as ended, removing the predictions and prefetches
 * of this session. <p>
 * Created (07.01.2002 15:12:07) by Christian Siefkes.<p>
 *
 * @param sessionID the ID of the current session (must be unique among all
 * sessions parallel in progress)
 */
synchronized void sessionEnded(final Integer sessionID) {
    // delete the predictions/prefetchesBySession of this session
    clearSession(sessionID);
} // method
} // class
