/*
 * Copyright (C) 2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.textgen;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import static net.siefkes.nlstego.NLStego.LOG;
import net.siefkes.nlstego.predict.PredictedToken;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * Base class that contains common code shared by the
 * {@link net.siefkes.nlstego.textgen.StegoEncoder} and
 * {@link net.siefkes.nlstego.textgen.StegoDecoder} classes.
 *
 * <p>Instances of this class are not stateless and must be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.12 $, $Date: 2005/07/28 15:03:12 $, $Author: siefkes $
 */
public class StegoCoder {

    /**
     * The text model used by this instance.
     */
    private final TextModel textModel;


    /**
     * Creates a new instance.
     *
     * @param conf the configuration to use
     * @param model the text model to use
     */
    public StegoCoder(final Configuration conf, final TextModel model) {
        super();
        textModel = model;
    }


    /**
     * Adds a debug statement that allows tracing the encoding/decoding of bits.
     * This method should only be called if
     * {@linkplain org.apache.commons.logging.impl.SimpleLog#isDebugEnabled()
     * debugging is enabled} for {@link net.siefkes.nlstego.NLStego#LOG}.
     *
     * @param num0 the number of '0' bits
     * @param added1 whether a '1' bit followed
     * @param predIndex the index of the current prediction
     * @param predCount the number of active predictions
     * @param token the predicted token
     */
    protected void debugBits(final int num0, final boolean added1,
            final int predIndex, final int predCount, final String token) {
        final StringBuilder msg = new StringBuilder();
        msg.append(num0).append("x0");
        if (added1) {
            msg.append("+1");
        }
        msg.append("|").append(predIndex + 1).append("/").append(predCount);
        msg.append("|").append(token);
        LOG.debug(msg);
    }

    /**
     * Adds a debug statement that allows tracing the occurrence of
     * single-element prediction lists. This method should only be called if
     * {@linkplain org.apache.commons.logging.impl.SimpleLog#isDebugEnabled()
     * debugging is enabled} for {@link net.siefkes.nlstego.NLStego#LOG}.
     *
     * @param token the predicted token
     */
    protected void debugSingleton(final String token) {
        LOG.debug("-|1/1|" + token);
    }

    /**
     * This method is called during encoding and decoding to filter unsuitable
     * predictions for the very first token in a stego text. It discards any
     * predicted tokens <em>not</em> {@link TextModel#startsWithWS(String)
     * starting with whitespace} to avoid confusion between tokens with and
     * without initial whitespace.
     *
     * @param predictions the original list of likely predictions
     * @return a filters list of predictions suitable for the very first token
     */
    protected List<PredictedToken> filterFirst(
            final List<PredictedToken> predictions) {
        final List<PredictedToken> result =
            new ArrayList<PredictedToken>(predictions.size());
        final Iterator<PredictedToken> predIter = predictions.iterator();
        PredictedToken pred;

        while (predIter.hasNext()) {
            pred = predIter.next();

            // only tokens that start with whitespace are ok as first token
            if (textModel.startsWithWS(pred.getToken())) {
                result.add(pred);
            } else {
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Ignoring as first token since it doesn't start "
                            + "with whitespace: " + pred.getToken());
                }
            }
        }

        if (LOG.isDebugEnabled() && (result.size() != predictions.size())) {
            LOG.debug("Filtered first predictions: " + result.toString()
                    + " instead of " + predictions.toString());
        }

        return result;
    }

    /**
     * Returns the text model used by this instance.
     *
     * @return the value of the attribute
     */
    protected TextModel getTextModel() {
        return textModel;
    }

    /**
     * Resets the state of this instance, to start encoding/decoding a new text
     * sequence.
     */
    protected void resetState() {
        textModel.startTokenSequence();
    }

    /**
     * Updates the state of a text model after choosing a token.
     *
     * @param token the chosen token
     */
    protected void updateState(final String token) {
        // update text model, without training
        textModel.updateState(token, false);
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("textModel", textModel)
            .toString();
    }

}
