/*
 * Copyright (C) 2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.textgen;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.Reader;
import java.text.MessageFormat;
import java.util.List;

import org.apache.commons.configuration.Configuration;

import static net.siefkes.nlstego.NLStego.LOG;
import static net.siefkes.nlstego.NLStego.RES;
import net.siefkes.nlstego.predict.PredictedToken;
import net.siefkes.nlstego.util.BitArray;
import net.siefkes.nlstego.util.InOutUtils;
import net.siefkes.nlstego.util.NumUtils;
import net.siefkes.nlstego.util.TextTokenizer;

/**
 * A stego decoder recovers the bit sequences hidden in stego texts.
 * Instances of this class are immutable and thus thread-safe.
 *
 * <p>Instances of this class are not stateless and must be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.16 $, $Date: 2005/07/12 17:02:20 $, $Author: siefkes $
 */
public class StegoDecoder extends StegoCoder {


    /**
     * Creates a new instance.
     *
     * @param conf the configuration to use
     * @param model the text model to use
     */
    public StegoDecoder(final Configuration conf, final TextModel model) {
        super(conf, model);
    }


    /**
     * Adapts the very first token of a stego, prepending the whitespace
     * that is assumed at the begin of each text.
     *
     * @param token the first token of a text
     * @return the adapted token with whitespace prepended
     */
    private String adaptFirst(final String token) {
        return TextTokenizer.DEFAULT_NORMALIZED_WS + token;
    }

    /**
     * Reads a stego text and decodes the embedded bit sequence.
     *
     * @param reader reader containing the stego text to decode; not closed by
     * this method
     * @return outStream the decoded bit sequence is written to this stream
     * @throws IOException if an I/O error occurs
     */
    public byte[] decode(final Reader reader) throws IOException {
        // prepare token sequence and read input
        resetState();
        final String text = InOutUtils.readToString(reader).trim();
        if (LOG.isDebugEnabled()) {
            LOG.debug("Text to decode: " + text);
        }

        final TextTokenizer tokenizer = getTextModel().createTokenizer(text);
        String token;
        boolean isFirst = true;
        boolean foundEndOfVInt = false;
        int decodedBits = 0;

        // preliminary array to decode the length VInt
        final BitArray vIntBits = new BitArray(32);
        int checkedBytes = 0;
        int completedBytes;

        // decode VInt from stego tokens
        while (!foundEndOfVInt && ((token = tokenizer.nextToken()) != null)) {
            decodedBits += decodeToken(vIntBits, token, Integer.MAX_VALUE,
                    isFirst);
            if (isFirst) {
                // disable after very first call
                isFirst = false;
            }
            completedBytes = decodedBits / 8;

            while ((completedBytes > checkedBytes) && !foundEndOfVInt) {
                // check if last completed byte ends the vInt
                foundEndOfVInt =
                    NumUtils.endsVInt(vIntBits.getByte(checkedBytes++));
            }
        }

        // decode byte length and create suitable bit array to store bit stream
        final int byteLength = (int)
            NumUtils.decodeVInt(new ByteArrayInputStream(vIntBits.getBytes()));
        final BitArray outBits = new BitArray(byteLength);
        int remainingBits = byteLength * 8;

        // copy extras bits to new bit array
        for (int i = checkedBytes * 8; i < decodedBits; i++) {
            outBits.setNext(vIntBits.get(i / 8, i % 8));
            remainingBits--;
        }

        // decode remaining bits
        while ((remainingBits > 0)
                && ((token = tokenizer.nextToken()) != null)) {
            remainingBits -= decodeToken(outBits, token, remainingBits,
                    isFirst);
        }

        if (remainingBits > 0) {
            // input ended early
            LOG.error(MessageFormat.format(RES.getString(
                    "dec/premature-end.error"), remainingBits));
        } else if (remainingBits < 0) {
            // not supposed to happen
            throw new RuntimeException("Implementation error: decoded "
                    + (-remainingBits) + " bits more than expected");
        }

        // return decoded bits
        return outBits.getBytes();
    }

    /**
     * Helper method that decodes the bits encoded in a stego token (if any).
     *
     * @param bitArray the decoded bits (if any) are appended to this bit array
     * @param token the token to decode
     * @param maxBits the maximum number of bits to decode
     * @param isFirst whether this is the very first token of a text
     * @return the number of decoded bits
     */
    private int decodeToken(final BitArray bitArray, final String token,
            final int maxBits, final boolean isFirst) {
        // get likely predictions for current state
        List<PredictedToken> predictions =
            getTextModel().predictLikelyWithoutRepetitions();
        final String actualToken;
        if (isFirst) {
            // adapt token remove unsuitable predictions if very first token
            predictions = filterFirst(predictions);
            actualToken = adaptFirst(token);
        } else {
            actualToken = token;
        }

        int tokenIndex = -1;
        final int result;

        // find token in list of prediction
        for (int i = 0; (i < predictions.size()) && (tokenIndex < 0); i++) {
            if (actualToken.equals(predictions.get(i).getToken())) {
                // got it
                tokenIndex = i;
/*                LOG.debug("Token '" + actualToken + "' is " + i
                        + "-th prediction"); */
            }
        }

        if (tokenIndex >= 0) {
            final int predictionCount = predictions.size();

            if (predictionCount > 1) {
                // index of last possible prediction
                final int lastIndex = Math.min(predictionCount - 1, maxBits);

                // decode bits and set result accordingly
                if (tokenIndex == lastIndex) {
                    // last prediction: append min(predictionCount-1,maxBits) 0s
                    result = lastIndex;

                    for (int i = 0; i < result; i++) {
                        bitArray.setNext(false);
                    }
                    if (LOG.isDebugEnabled()) {
                        debugBits(result, false, tokenIndex, predictionCount,
                                actualToken);
                    }
                } else {
                    // i-th prediction: append i 0s followed by 1
                    for (int i = 0; i < tokenIndex; i++) {
                        bitArray.setNext(false);
                    }

                    bitArray.setNext(true);
                    result = tokenIndex + 1;

                    if (LOG.isDebugEnabled()) {
                        debugBits(tokenIndex, true, tokenIndex, predictionCount,
                                actualToken);
                    }
                }
            } else {
                // no choice among predictions -- nothing to decode
                result = 0;
                if (LOG.isDebugEnabled()) {
                    debugSingleton(actualToken);
                }
            }

            updateState(actualToken);
        } else {
            // token not found -- this is not necessarily an error since the
            // extended model allows inserting arbitrary unpredicted tokens
            // (to avoid non-reversibility attack)
            LOG.info(MessageFormat.format(
                    RES.getString("dec/unknown-token.info"), actualToken));
            if (LOG.isDebugEnabled()) {
                LOG.debug("Expected tokens: " + predictions);
            }
            result = 0;
        }

        return result;
    }

}
