/*
 * Copyright (C) 2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.textgen;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.ArrayUtils;

import static net.siefkes.nlstego.NLStego.LOG;
import static net.siefkes.nlstego.NLStego.RES;
import net.siefkes.nlstego.predict.PredictedToken;
import net.siefkes.nlstego.util.BitArray;
import net.siefkes.nlstego.util.NumUtils;

/**
 * A stego encoder hides bit sequences into inconspicuous stego texts.
 * Instances of this class are immutable and thus thread-safe.
 *
 * <p>Instances of this class are not stateless and must be synchronized
 * externally, if required.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.20 $, $Date: 2005/07/28 15:03:12 $, $Author: siefkes $
 */
public class StegoEncoder extends StegoCoder {

    /**
     * Creates a new instance.
     *
     * @param conf the configuration to use
     * @param model the text model to use
     */
    public StegoEncoder(final Configuration conf, final TextModel model) {
        super(conf, model);
    }

    /**
     * Adapts the very first token of a stego, removing the initial whitespace
     * that is implicit at the begin of each text.
     *
     * @param token the first token of a text
     * @return the adapted token with whitespace removed
     */
    private String adaptFirst(final String token) {
        // ensure that the token is a valid first token (starts with whitespace)
        if (getTextModel().startsWithWS(token)) {
            // delete initial space char
            return token.substring(1);
        } else {
            // not supposed to happen
            throw new RuntimeException("Implementation error: '" + token
                + "' is not a valid first token (must start with whitespace)");
        }
    }

    /**
     * Reads a bit sequence and encodes it into an inconspicuous stego text.
     *
     * @param in stream containing the bit sequence to encode; not closed
     * by this method
     * @return the generated stego text
     * @throws IOException if an I/O error occurs
     */
    public String encode(final InputStream in)
    throws IOException {
        return encode(IOUtils.toByteArray(in));
    }

    /**
     * Reads a bit sequence and encodes it into an inconspicuous stego text.
     *
     * @param bytes the bit sequence to encode
     * @return the generated stego text
     */
    public String encode(final byte[] bytes) {
        // prepare token sequence and init result
        resetState();
        final StringBuilder result = new StringBuilder();
        boolean isFirst = true;

        // prepend length in bytes as VInt
        final byte[] lengthInBytes = NumUtils.encodeVInt(bytes.length);
        final byte[] joinedBytes = ArrayUtils.addAll(lengthInBytes, bytes);

        final BitArray inBits = new BitArray(joinedBytes);
        boolean bit;
        List<PredictedToken> predictions;
        int predictionCount;
        int predIndex;
        PredictedToken chosenPred;
        String chosenToken = null;
        boolean moreBits = true;
        int bitCount;

        // process all bits
        while (moreBits) {
            // get likely predictions for current state
            predictions = getTextModel().predictLikelyWithoutRepetitions();
            if (isFirst) {
                // remove unsuitable tokens if very first token
                predictions = filterFirst(predictions);
            }
            predictionCount = predictions.size();
            bit = false;
            bitCount = 0;

            if (predictionCount > 1) {
                // read up to n-1 bits until next 1 or end-of-stream
                while (bitCount < (predictionCount - 1) && !bit
                        && (moreBits = inBits.hasNext())) { // == true
                    bit = inBits.next();
                    bitCount++;
                }

                if (bit) {
                    // (bitCount-1) 0 bits followed by 1
                    predIndex = bitCount - 1;
                    chosenPred = predictions.get(predIndex);

                    if (LOG.isDebugEnabled()) {
                        debugBits(predIndex, true, predIndex, predictionCount,
                                chosenPred.getToken());
                    }
                } else {
                    // all m bits are 0: append m-th prediction
                    predIndex = bitCount;
                    chosenPred = predictions.get(predIndex);

                    if (LOG.isDebugEnabled()) {
                        debugBits(predIndex, false, predIndex, predictionCount,
                                chosenPred.getToken());
                    }
                }
            } else {
                if (predictionCount == 1) {
                    // only a single prediction: append w/o encoding anything
                    chosenPred = predictions.get(0);
                    if (LOG.isDebugEnabled()) {
                        debugSingleton(chosenPred.getToken());
                    }
                } else {
                    // not supposed to happen
                    throw new IllegalStateException(RES.getString(
                            "enc/no-predictions"));
                }
            }

            // store chosen prediction + update model state
            chosenToken = chosenPred.getToken();
            updateState(chosenToken);

            if (isFirst) {
                // adapt very first token of a text
                chosenToken = adaptFirst(chosenToken);
                isFirst = false;
            }

            // append token and check if it's necessary to go on
            result.append(chosenToken);
            moreBits = inBits.hasNext();
        }

        // append up to 100 random words until suitable end-of-sentence token
        if (!getTextModel().isEndOfSentenceToken(chosenToken)) {
            getTextModel().predictTokens(0, 100, false, result);
        }

        return result.toString();
    }

}
