/*
 * Copyright (C) 2003-2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.textgen;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.URL;
import java.net.URLConnection;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import net.siefkes.nlstego.NLStego;
import static net.siefkes.nlstego.NLStego.LOG;
import static net.siefkes.nlstego.NLStego.RES;
import net.siefkes.nlstego.predict.PPMFetchStats;
import net.siefkes.nlstego.predict.PredictedToken;
import net.siefkes.nlstego.predict.PredictionManager;
import net.siefkes.nlstego.predict.StatelessFilter;
import net.siefkes.nlstego.util.ContentType;
import net.siefkes.nlstego.util.InOutUtils;
import net.siefkes.nlstego.util.TextTokenizer;
import net.siefkes.nlstego.util.TokenizerFactory;
import net.siefkes.nlstego.util.Util;

import org.apache.commons.collections.buffer.CircularFifoBuffer;
import org.apache.commons.collections.map.LRUMap;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.vfs.FileContent;


/**
 * A text model that can be used to complete texts or to generate "typical"
 * texts based a prediction model.
 * One or several sample texts must be provided to train the prediction
 * models ({@link #train(CharSequence)}), then the expected next token(s)
 * in a sequence can be predicted
 * ({@link #predictTokens(int, int, boolean, StringBuilder)}) or "typical"
 * texts can be generated.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.30 $, $Date: 2005/07/28 15:03:12 $, $Author: siefkes $
 */
public class TextModel implements StatelessFilter {

    /**
     * Dummy object used for the {@link #lastTokens} buffer.
     */
    private static final Object DUMMY = new Object();


    /**
     * Pseudo-token to use when a token is expected but none is available
     * (e.g., before the first token starting a text).
     */
    public static final String VOID_TOKEN = "";

    /**
     * The maximum ratio of words generated extra when trying to find a
     * suitable ending.
     */
    private static final int DEFAULT_EXTRA_RATE = 3;

    /**
     * The prediction manager that combined predictions from different
     * models.
     */
    private final PredictionManager predMan;

    /**
     * A list of tokens that can end a text.
     */
    private final List finalTokens;

    /**
     * Threshold used for discarding unlikely predictions.
     */
    private final double threshold;

    /**
     * Random number generator for chosing among predictions.
     */
    private Random random = null; // lazy initialization

    /**
     * A buffer of the last generated/read tokens, used to avoid repetitious
     * phrases.
     */
    private CircularFifoBuffer lastTokens = null; // lazy initialization

    /**
     * A cache of recent phrases generated/read by this model, used to avoid
     * repetitous phrases.
     */
    private LRUMap recentPhrases = null; // lazy initialization

    /**
     * Creates a new instance.
     *
     * @param conf the configuration to use
     */
    public TextModel(final Configuration conf) {
        super();

        // initialize members
        finalTokens = conf.getList("textgen.finaltokens",
                Arrays.asList(".", "!", "?"));
        threshold = conf.getDouble("textgen.prob.threshold", 4.0);
        predMan = new PredictionManager(conf);

        // check config param
        if (threshold <= 1.0) {
            throw new IllegalArgumentException(MessageFormat.format(
                RES.getString("model/too-small-threshold"), threshold));
        }

        // for now just use a single standard model
        predMan.registerPredictionModel(new PPMFetchStats(conf, this));
    }


    /**
     * {@inheritDoc}
     * This implementation rejects tokens that start with an alphanumeric
     * character to ensure correct tokenization when decoding.
     */
    public boolean acceptForStateless(final String key) {
        if (key.length() > 0) {
            return !Character.isLetterOrDigit(key.charAt(0));
        } else {
            // not supposed to happen (empty token)
            return true;
        }
    }

    /**
     * Creates a tokenizer for splitting text into texts.
     *
     * @param contents the text to tokenize
     * @return a tokenizer initialized to the given text
     */
    public TextTokenizer createTokenizer(final CharSequence contents) {
        final TextTokenizer result =
            TokenizerFactory.createThoroughTokenizer(contents);
        result.setNormalizedWhitespacePrepended(true);
        return result;
    }

    /**
     * Helper method that flattens the last tokens buffer into a string.
     * @return the sequence of last tokens
     */
    private String flattenLastTokens() {
        final StringBuilder result = new StringBuilder();
        final Iterator lastTokensIter = lastTokens.iterator();

        while (lastTokensIter.hasNext()) {
            result.append((String) lastTokensIter.next());
        }

        return result.toString();
    }

    /**
     * Returns the threshold for discarding unlikely predictions.
     *
     * @return the value of the attribute
     */
    public double getThreshold() {
        return threshold;
    }

    /**
     * Checks whether a token is suitable as the end of a sentence.
     *
     * @param token the token to check
     * @return <code>true</code> iff this token is a typical end-of-sentence
     * token
     */
    public boolean isEndOfSentenceToken(final String token) {
        return finalTokens.contains(token);
    }

    /**
     * Finds the first (most likely) {@linkplain #isEndOfSentenceToken(String) 
     * end-of-sentence token} among a list of predictions.
     *
     * @param predictions the predictions to check
     * @return the first (most likely} end-of-sentence token found in the
     * list; or <code>null</code> if there are none
     */
    public PredictedToken findFirstEndOfSentenceToken(
            final List<PredictedToken> predictions) {
        final Iterator<PredictedToken> predIter = predictions.iterator();
        PredictedToken pred;
        String token;

        while (predIter.hasNext()) {
            pred = predIter.next();
            token = pred.getToken();

            if (isEndOfSentenceToken(token)) {
                return pred; // got it
            }
        }

        return null; // no match
    }

    /**
     * Returns a list of likely predictions, sorted by probability. Applies the
     * configured dynamic {@linkplain #getThreshold() threshold} to discard
     * unlikely predictions:
     * if <em>P(pred<sub>n+1</sub>) < P(pred<sub>n</sub>) / threshold</em>
     * <em>pred<sub>n+1</sub></em> and all following (less likely) predictions
     * will be discarded.
     *
     * <p>Don't forget to call {@link #updateState(String, boolean)} after
     * choosing a token from the result list.
     *
     * @return a list of predictions whose probability is high enough, sorted
     * by probability
     */
    public List<PredictedToken> predictLikely() {
        final List<PredictedToken> predTokenList =
            predMan.choosePredictions();
        final Iterator<PredictedToken> predTokenIter = predTokenList.iterator();
        final List<PredictedToken> result =
            new ArrayList<PredictedToken>(predTokenList.size());
        PredictedToken lastPred;
        PredictedToken currentPred = null;
        boolean goOn = true;
//        boolean log = false; // debug

        while (goOn && predTokenIter.hasNext()) {
            lastPred = currentPred;
            currentPred = predTokenIter.next();

            if ((lastPred == null) || (currentPred.getProbability()
                    >= (lastPred.getProbability() / threshold))) {
                result.add(currentPred);
/*                if (currentPred.getToken().equals(" and")) {
                    log = true;
                } */
            } else {
                goOn = false;
/*                LOG.info("Ignoring prediction " + currentPred
                        + " since it's less probably than previous prediction "
                        + lastPred + " / threshold " + threshold); */
            }
        }
/*        if (log) {
            LOG.info("All: " + predTokenList + "; likely: " + result);
        } */

        return result;
    }

    /**
     * Returns a list of likely predictions, sorted by probability. Predictions
     * that would lead to the repetition of a recent phrase are filtered out.
     *
     * @return a list of likely predictions
     */
    public List<PredictedToken> predictLikelyWithoutRepetitions() {
        if (lastTokens == null) {
            // lazily init fields on first call to this method
            lastTokens = new CircularFifoBuffer(4);
            recentPhrases = new LRUMap(100);
        }

        final List<PredictedToken> predTokenList = predictLikely();

        // filter against phrase cache
        final List<PredictedToken> result =
            new ArrayList<PredictedToken>(predTokenList.size());
        final Iterator<PredictedToken> predTokenIter = predTokenList.iterator();
        final String flattenedLastTokens = flattenLastTokens();
        PredictedToken pred;
        String tokenSequence;

        while (predTokenIter.hasNext()) {
            pred = predTokenIter.next();
            tokenSequence = flattenedLastTokens + pred.getToken();

            if (!recentPhrases.containsKey(tokenSequence)) {
                result.add(pred);
            } else {
                // otherwise the prediction would complete a recent phrase
                if (LOG.isDebugEnabled()) {
                    LOG.debug("Avoiding recent phrase: " + tokenSequence);
                }
            }
        }

        if (!result.isEmpty()) {
            return result;
        } else {
            if (LOG.isDebugEnabled()) {
                LOG.debug("Returning full list of likely predictions since all "
                        + "are in recent phrase cache");
            }
            return predTokenList;
        }
    }

    /**
     * Predicts the tokens that will probably be generated next, randomly
     * selecting tokens according to their probability (considering only the
     * first 50% of the probability mass to avoid unlikely predictions which
     * are likely to turn out dead ends). The number
     * of predicted words will be in the range from <code>number</code> to
     * <code>4/3 number</code> to attain a suitable ending (it might also be
     * lower if the prediction model ran out of predictions).
     *
     * @param number the number of words to generate, >0
     * @param reset whether to reset the model to the start of a new sentence;
     * if <code>false</code>, the current model state won't be modified
     * @param appender a StringBuilder to append the predictions to
     */
    public final void predictTokens(final int number, final boolean reset,
            final StringBuilder appender) {
        final int maxExtra = number / DEFAULT_EXTRA_RATE;
        predictTokens(number, maxExtra, reset, appender);
    }

    /**
     * Predicts the tokens that will probably be requested next, randomly
     * selecting tokens according to their probability (considering only the
     * first 50% of the probability mass to avoid unlikely predictions which
     * are likely to turn out dead ends). The number
     * of predicted words will be in the range from <code>number</code> to
     * <code>number + maxExtra</code> to attain a suitable ending (it might also
     * be lower if the prediction model ran out of predictions).
     *
     * @param number the number of words to generate, >0
     * @param maxExtra the maximum number of extra words that can be
     * generated until a suitable ending is found
     * @param reset whether to reset the model to the start of a new sentence;
     * if <code>false</code>, the current model state won't be modified
     * @param appender a StringBuilder to append the predictions to
     * (for output/debugging)
     */
    public final void predictTokens(final int number, final int maxExtra,
            final boolean reset, final StringBuilder appender) {
        List<PredictedToken> predTokenList;
        PredictedToken predToken = null;
        String predTokenValue;
        boolean goOn = true;
        int generated = 0;
        boolean enoughGenerated = false;
        final int maxNumber = number + maxExtra;
        // immediately stop if finding and end-of-sentence token after
        // generating half the extra words
        final int criticalNumber = number + (maxExtra / 2);

        // lazy initialization of random generator
        if (random == null) {
            random = new Random();
        }

        if (reset) {
            // reset model to start of new sentence
            startTokenSequence();
        }

        while (goOn) {
            predTokenList = predictLikely();

            if (predTokenList.size() >= 1) {
                if (generated >= criticalNumber) {
                    // immediately stop if finding and end-of-sentence token
                    // after generating half the extra words
                    predToken = findFirstEndOfSentenceToken(predTokenList);
                } else {
                    predToken = null;
                }

                if (predToken == null) {
                    // randomly choose a prediction
                    final double randNum = random.nextDouble();
                    final Iterator<PredictedToken> predTokenIter =
                        predTokenList.iterator();
                    double summedProb = 0.0;
                    boolean gotIt = false;

                    while (predTokenIter.hasNext() && !gotIt) {
                        predToken = predTokenIter.next();
                        summedProb += predToken.getProbability();

                        if (summedProb >= randNum) {
                            // select token as soon as summed probability mass
                            // reaches the random number
                            gotIt = true;
/*                            Util.LOG.info("Using token " + i +  "/"
                                    + predTokenList.size() + ": " + predToken
                                    + ", summed prob.: " + summedProb
                                    + ", random number: " + randNum); */
                        }
                    }

                    if (!gotIt) {
                        // random number is higher than summed prob. of all
                        // tokens (since there are some ignored predictions):
                        // use most likely token
                        predToken = predTokenList.get(0); 
/*                      Util.LOG.info("Using first token of "
                              + predTokenList.size() + " since summed prob. "
                              + summedProb + " < random number " + randNum); */
                    }
                } else {
                    if (LOG.isDebugEnabled()) {
                        LOG.debug("Using most likely end-of-sentence token: "
                                + predToken);
                    }
                }

                generated++;
                predTokenValue = predToken.getToken();
                appender.append(predTokenValue);
                // feed the token back for updating the model state
                updateState(predTokenValue, false);

                if ((!enoughGenerated) && (generated >= number)) {
                    // generated the required number of words
                    enoughGenerated = true;
                }

                if (enoughGenerated) {
                    if (generated >= maxNumber) {
                        // reached allowed maximum, so we have to stop
                        goOn = false;
//                      Util.LOG.info("Reached allowed maximum: " + maxNumber);
                    } else {
                        // check whether the last tokens is approved by the
                        // final token list
                        if (isEndOfSentenceToken(predTokenValue)) {
                            // that's a fine ending--we're done!
                            goOn = false;
//                          Util.LOG.info("Last token is in final token cache");
                        }
                    }
                }
            } else {
                // nothing to predict---giving up
                appender.append(" ???");
                   goOn = false;
            }
        }
    }

    /**
     * Checks whether a token starts with
     * {@link TextTokenizer#DEFAULT_NORMALIZED_WS normalized whitespace}.
     *
     * @param token the token to check
     * @return <code>true</code> if the token starts with normalized whitespace
     */
    protected boolean startsWithWS(final String token) {
        return (token.length() > 0)
                && (token.charAt(0) == TextTokenizer.DEFAULT_NORMALIZED_WS);
    }

    /**
     * Prepares the model to start a new token sequence.
     */
    public void startTokenSequence() {
        // use first final token as starting point
        final String preStartToken = (String) finalTokens.get(0);
        updateState(preStartToken, false);

        if (lastTokens != null) {
            lastTokens.clear();
            recentPhrases.clear();
        }

        if (LOG.isDebugEnabled()) {
            // print model stats
            LOG.debug("Calculating model stats...");
            LOG.debug(predMan.printModelStats());
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("prediction manager", predMan)
            .append("final tokens", finalTokens)
            .append("threshold", threshold)
            .toString();
    }

    /**
     * Trains the prediction models from the provided string.
     *
     * @param contents the training data
     */
    public final void train(final CharSequence contents) {
        final TextTokenizer tokenizer = createTokenizer(contents);
        String normalizedToken;

        while ((normalizedToken = tokenizer.nextToken()) != null) {
            updateState(normalizedToken, true);
        }
    }

    /**
     * Trains the prediction models from the contents of a file.
     *
     * @param file the source of the training data
     * @throws IOException if an I/O error occurs
     */
    public final void train(final File file) throws IOException {
        final Reader reader = new FileReader(file);
        try {
            train(reader);
        } finally {
            InOutUtils.tryToClose(reader);
        }
    }

    /**
     * Trains the prediction models from the contents of the provided reader.
     * The reader is <em>not</em> closed after usage.
     *
     * @param reader the source of the training data
     * @throws IOException if an I/O error occurs
     */
    public final void train(final Reader reader) throws IOException {
        final String contents = InOutUtils.readToString(reader);
        train(contents);
    }

    /**
     * Trains the prediction models from an array of file paths, using the
     * {@link InOutUtils#listFileContents(String[], boolean)} method to resolve
     * file contents.
     *
     * @param filePaths the array of file paths
     * @param recurse whether to recursively add the children of folders and
     * other files containing nested entries
     * @throws IOException if an I/O error occurs
     */
    public final void train(final String[] filePaths, final boolean recurse)
    throws IOException {
        final FileContent[] contents =
            InOutUtils.listFileContents(filePaths, recurse);
        Reader reader = null;

        for (int i = 0; i < contents.length; i++) {
            try {
                reader = InOutUtils.openReader(contents[i]);
                train(reader);
            } finally {
                InOutUtils.tryToClose(reader);
            }
        }
    }

    /**
     * Trains the prediction models from the contents of an URL.
     *
     * @param url the source of the training data
     * @throws IOException if an I/O error occurs
     */
    public final void train(final URL url) throws IOException {
        final URLConnection urlConn = url.openConnection();
        final InputStream in = urlConn.getInputStream();
        final ContentType contentType =
            ContentType.determineContentType(urlConn.getContentType());
        final Reader reader;

        if (contentType.getCharset() != null) {
            // use specified charset
            reader = new InputStreamReader(in, contentType.getCharset());
        } else {
            // use default charset
            reader = new InputStreamReader(in);
        }

        try {
            train(reader);
        } finally {
            InOutUtils.tryToClose(reader);
        }
    }

    /**
     * Updates the prediction models by feeding them the specified token.
     *
     * @param token the token to add
     * @param train <code>true</code> iff the models should be trained, i.e.
     * this request should be incorporated into the prediction models;
     * otherwise only the internal state of the models will be changed so the
     * next call for predictions will consider the updated state
     */
    public final void updateState(final String token, final boolean train) {
         // delegate for actual processing/training
        predMan.addItem(token, train);

        // update last tokens + recent phrases, if used
        if (lastTokens != null) {
            final String flattenedLastTokens = flattenLastTokens();
            recentPhrases.put(flattenedLastTokens + token, DUMMY);
            lastTokens.add(token);
        }
    }


    /**
     * Does some tests, tokenizing and printing some files.
     *
     * @param args the command line arguments (ignored)
     */
    public static void main(final String[] args) {

        final String[] sampleURLs = {
            "http://www.ietf.org/rfc/rfc1945.txt", // HTTP 1.0 spec
            "http://www.ietf.org/rfc/rfc2616.txt", // HTTP 1.1 spec
            //"http://www.w3.org/TR/REC-xml", // XML spec
        };

        InputStreamReader reader = null;

        try {
            final Configuration defaultConfig =
                Util.loadConfiguration(NLStego.ID);
            final TextModel textGen = new TextModel(defaultConfig);
            URL url;
            URLConnection urlConn;
            InputStream is;
            String charset;

            for (int i = 0; i < sampleURLs.length; i++) {
                url = new URL(sampleURLs[i]);
                System.out.println(
                    "--------------------------------------------------");
                System.out.println("URL: " + url.toExternalForm());
                urlConn = url.openConnection();
                is = urlConn.getInputStream();
                final ContentType contentType =
                    ContentType.determineContentType(urlConn.getContentType());
                charset = contentType.getCharset();

                // Use the default encoding if none is specified
                if (charset == null) {
                    charset = InOutUtils.STANDARD_HTTP_CHARSET;
                }

                reader = new InputStreamReader(is, charset);
                textGen.train(reader);
                InOutUtils.tryToClose(reader);
            }
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            InOutUtils.tryToClose(reader);
        }
    }

}
