/*
 * Copyright (C) 2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.util;

import java.io.IOException;
import java.io.InputStream;
import java.text.MessageFormat;

import net.siefkes.nlstego.NLStego;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * This class provides read access to an array of bits (booleans).
 *
 * @author Christian Siefkes
 * @version $Revision: 1.9 $, $Date: 2005/07/12 17:02:22 $, $Author: siefkes $
 */
public class BitArray {

    /**
     * The index of the last (lowest-order) bit in a byte.
     */
    private static final byte LAST_BIT = 7;


    /**
     * The wrapped byte array.
     */
    private final byte[] bytes;

    /**
     * Current byte position in the buffer.
     */
    private int currentByte = 0;

    /**
     * Current bit position in the buffer.
     */
    private byte currentBit = 0;


    /**
     * Creates a new instance. Creates an internal byte array of the specified
     * length, initially setting all bytes (and thus all bits) to 0.
     *
     * @param byteLength the length of the internal array of bytes (in bytes)
     */
    public BitArray(final int byteLength) {
        super();
        // all bytes will be initialized to the default value 0
        bytes = new byte[byteLength];
    }

    /**
     * Creates a new instance.
     *
     * @param myBytes the array of bytes to wrap; any changes to this array
     * will immediately affect the wrapping bit array
     */
    public BitArray(final byte[] myBytes) {
        super();
        bytes = myBytes;
    }

    /**
     * Creates a new instance, reading the data to wrap from an input stream.
     *
     * @param in the contents of this input stream will be wrapped
     * @throws IOException if an I/O error occurs
     */
    public BitArray(final InputStream in) throws IOException {
        this(IOUtils.toByteArray(in));
    }


    /**
     * Returns the length of this array, in bytes.
     *
     * @return the length of this array, in bytes
     */
    public int byteLength() {
        return bytes.length;
    }

    /**
     * Helper method that ensures that the given indexes are valid, throwing
     * an exception if one of them is out of range.
     *
     * @param byteIndex the index of the byte to read from, must by in the range
     * from <code>0</code> to <code>{@link #byteLength()} - 1}</code>
     * @param bitIndex the index of the bit within the byte, must in in the
     * range from 0 (highest-order bit) to 7 (lowest-order bit)
     * @throws ArrayIndexOutOfBoundsException if one of the indexes is outside
     * the allowed range
     */
    private void checkArguments(final int byteIndex, final int bitIndex)
    throws ArrayIndexOutOfBoundsException {
        if ((byteIndex < 0) || byteIndex >= byteLength()) {
            throw new ArrayIndexOutOfBoundsException(MessageFormat.format(
                 NLStego.RES.getString("bits/illegal-byte-index"), byteIndex));
        }
        if ((bitIndex < 0) || bitIndex > LAST_BIT) {
            throw new ArrayIndexOutOfBoundsException(MessageFormat.format(
                 NLStego.RES.getString("bits/illegal-bit-index"), bitIndex));
        }
    }

    /**
     * Returns the value of a bit.
     *
     * @param byteIndex the index of the byte to read from, must by in the range
     * from <code>0</code> to <code>{@link #byteLength()} - 1}</code>
     * @param bitIndex the index of the bit within the byte, must in in the
     * range from 0 (highest-order bit) to 7 (lowest-order bit)
     * @return the value of the specified bit: <code>true</code> for 1,
     * <code>false</code> for 0
     * @throws ArrayIndexOutOfBoundsException if one of the indexes is outside
     * the allowed range
     */
    public boolean get(final int byteIndex, final int bitIndex)
    throws ArrayIndexOutOfBoundsException {
        checkArguments(byteIndex, bitIndex);
        final int bit = bytes[byteIndex] & (1 << (LAST_BIT - bitIndex));
        return (bit != 0);
    }

    /**
     * Returns the value of a byte.
     *
     * @param byteIndex the index of the byte to get, must by in the range
     * from <code>0</code> to <code>{@link #byteLength()} - 1}</code>
     * @return the value of the specified byte
     * @throws ArrayIndexOutOfBoundsException if the index is outside the
     * allowed range
     */
    public byte getByte(final int byteIndex)
    throws ArrayIndexOutOfBoundsException {
        return bytes[byteIndex];
    }

    /**
     * Returns the internal byte array wrapped by this instance. Note that this
     * method returns the actually used array (not a copy), so all changes to
     * the returned array will immediately influence the outer instance and vice
     * versa.
     *
     * @return the wrapped array
     */
    public byte[] getBytes() {
        return bytes;
    }

    /**
     * Returns the length of this array, in bits.
     *
     * @return the length of this array, in bits
     */
    public long length() {
        return 8L * byteLength();
    }

    /**
     * Checks whether there are more bits to read or set from this array. Iff
     * this method returns <code>false</code>, the next call to {@link #next()}
     * or {@link #setNext(boolean)} will throw an exception.
     *
     * @return <code>true</code> iff there are more bits to read or set
     */
    public boolean hasNext() {
        return (currentByte < byteLength());
    }

    /**
     * Returns the next unread or unset bit from this array.
     *
     * @return the value of the next unread bit: <code>true</code> for 1,
     * <code>false</code> for 0
     * @throws ArrayIndexOutOfBoundsException if all bits from this array
     * have already been read or set
     */
    public boolean next() throws ArrayIndexOutOfBoundsException {
        final boolean result = get(currentByte, currentBit);
        currentBit++;

        if (currentBit > LAST_BIT) {
            currentBit = 0;
            currentByte++;
        }

        return result;
    }

    /**
     * Modifies the value of a bit.
     *
     * @param byteIndex the index of the byte to modify, must by in the range
     * from <code>0</code> to <code>{@link #byteLength()} - 1}</code>
     * @param bitIndex the index of the bit within the byte, must in in the
     * range from 0 (highest-order bit) to 7 (lowest-order bit)
     * @param value the new value of the specified bit: <code>true</code> for 1,
     * <code>false</code> for 0
     * @throws ArrayIndexOutOfBoundsException if one of the indexes is outside
     * the allowed range
     */
    public void set(final int byteIndex, final int bitIndex,
            final boolean value)
    throws ArrayIndexOutOfBoundsException {
        checkArguments(byteIndex, bitIndex);
        final int bitPattern = 1 << (LAST_BIT - bitIndex);

        if (value) {
            // set bit to 1 w/o changing others
            bytes[byteIndex] |= bitPattern;
        } else {
            // set bit to 0 w/o changing others
            bytes[byteIndex] &= ~bitPattern;
        }
    }

    /**
     * Returns the next unread or unset bit from this array.
     *
     * @param value the new value of the bit: <code>true</code> for 1,
     * <code>false</code> for 0
     * @throws ArrayIndexOutOfBoundsException if all bits from this array
     * have already been read or set
     */
    public void setNext(final boolean value)
    throws ArrayIndexOutOfBoundsException {
        set(currentByte, currentBit, value);
        currentBit++;

        if (currentBit > LAST_BIT) {
            currentBit = 0;
            currentByte++;
        }
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("length in bits", length())
            .toString();
    }

}
