/*
 * Copyright (C) 2003-2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.util;

import java.text.MessageFormat;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.siefkes.nlstego.NLStego;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;


/**
 * A simple wrapper for the MIME type and the character set of a document.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.5 $, $Date: 2005/07/12 17:02:22 $, $Author: siefkes $
 */
public final class ContentType {

    /**
     * Context key recommended to store the MIME type string.
     */
    public static final String KEY_MIME_TYPE = "MIME-Type";

    /**
     * MIME type for HTML.
     */
    public static final String MIME_HTML = "text/html";

    /**
     * MIME type for plain text.
     */
    public static final String MIME_PLAIN = "text/plain";

    /**
     * The regex pattern for parsing the "Content-Type" header field of an URL.
     * The first capturing group captures the MIME type, the second one the
     * character set, if specified.
     */
    private static final Pattern CONTENT_TYPE_PATTERN = Pattern.compile(
        "([^\\s;]+)(?:.*?;\\s*charset\\s*=\\s*\"?([^\\s;\"]*)\"?\\s*;?)?",
        Pattern.CASE_INSENSITIVE);

    /**
     * Factory method to determines the content type based an "Content-Type"
     * header. The "Content-Type" header is
     * examined first, if it is specified (not null or empty). If this does
     * not lead to a result, the extension is examined.
     *
     * @param contentTypeHeader the "Content-Type" header of an URL, as returned
     * by {@link java.net.URLConnection#getContentType()} (might be
     * <code>null</code>)
     * @return the determined content type
     */
    public static ContentType determineContentType(
            final String contentTypeHeader) {
        String type = null;
        String charSet = null;

        // parse Content-Type header, if given
        if (StringUtils.isNotEmpty(contentTypeHeader)) {
            final Matcher contentTypeMatcher =
                CONTENT_TYPE_PATTERN.matcher(contentTypeHeader);
            if (contentTypeMatcher.find()) {
                type = contentTypeMatcher.group(1);
                charSet = contentTypeMatcher.group(2);
            } else {
                // otherwise it's not a valid MIME type
                NLStego.LOG.warn(MessageFormat.format(NLStego.RES.getString(
                        "contenttype/invalid.warn"), contentTypeHeader));
            }
        }

        return new ContentType(type, charSet);
    }

    /**
     * The character set.
     */
    private final String charset;

    /**
     * The MIME type, <code>null</code> if unknown.
     */
    private final String mimeType;

    /**
     * Non-public constructor for internal use. Use the factory methods instead.
     *
     * @param type the MIME type of this document, <code>null</code> if unknown;
     * will be converted to lower-case
     * @param characterSet the character set of this document,
     * <code>null</code> if unknown
     */
    private ContentType(final String type, final String characterSet) {
        super();
        mimeType = StringUtils.lowerCase(type);
        charset = characterSet;
    }

    /**
     * Returns the character set.
     *
     * @return the character set, or <code>null</code> if unknown
     */
    public String getCharset() {
        return charset;
    }

    /**
     * Returns the MIME type.
     *
     * @return the MIME type, or <code>null</code> if unknown
     */
    public String getMimeType() {
        return mimeType;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a textual representation
     */
    public String toString() {
        return new ToStringBuilder(this)
            .append("mime type", mimeType)
            .append("charset", charset)
            .toString();
    }

}
