/*
 * Copyright (C) 2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.util;

import java.util.Map;

import org.apache.commons.collections.map.LRUMap;

/**
 * An extension of the LRUMap that also allows transparent (invisible) read
 * operations which do not modify the access order of the map entries.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2005/07/12 17:02:22 $, $Author: siefkes $
 */
public class ExtendedLRUMap extends LRUMap {

    /**
     * Creates a new instance.
     */
    public ExtendedLRUMap() {
        super();
    }

    /**
     * Creates a new instance.
     * 
     * @param maxSize the maximum size of the map
     */
    public ExtendedLRUMap(final int maxSize) {
        super(maxSize);
    }

    /**
     * Creates a new instance.
     * 
     * @param maxSize the maximum size of the map
     * @param scanUntilRemovable scan until a removeable entry is found, default
     * <code>false</code>
     */
    public ExtendedLRUMap(final int maxSize, final boolean scanUntilRemovable) {
        super(maxSize, scanUntilRemovable);
    }

    /**
     * Creates a new instance.
     * 
     * @param maxSize the maximum size of the map
     * @param loadFactor the load factor
     */
    public ExtendedLRUMap(final int maxSize, final float loadFactor) {
        super(maxSize, loadFactor);
    }

    /**
     * Creates a new instance.
     * 
     * @param maxSize the maximum size of the map
     * @param loadFactor the load factor
     * @param scanUntilRemovable scan until a removeable entry is found, default
     * <code>false</code>
     */
    public ExtendedLRUMap(final int maxSize, final float loadFactor,
            final boolean scanUntilRemovable) {
        super(maxSize, loadFactor, scanUntilRemovable);
    }

    /**
     * Constructor copying elements from another map.
     * The maximum size is set from the map's size.
     *
     * @param map the map to copy
     */
    public ExtendedLRUMap(final Map map) {
        super(map);
    }

    /**
     * Constructor copying elements from another map.
     * The maximum size is set from the map's size.
     *
     * @param map the map to copy
     * @param scanUntilRemovable scan until a removeable entry is found, default
     * <code>false</code>
     */
    public ExtendedLRUMap(final Map map, final boolean scanUntilRemovable) {
        super(map, scanUntilRemovable);
    }


    /***
     * Returns the value to the specified key, without changing the LRU
     * structure. Use this instead of <code>get</code>, if you want to make
     * an "invisible" request that is ignored by the replacement algorithm.
     *
     * @param key the key to look up
     * @return the mapped value, <code>null</code> if no match
     */
    public Object transparentGet(final Object key) {
        LinkEntry entry = (LinkEntry) getEntry(key);
        return (entry == null) ? null : entry.getValue();
    }

}
