/*
 * Copyright (C) 2003-2005 Christian Siefkes <christian@siefkes.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.util;

import java.io.ByteArrayInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import net.siefkes.nlstego.NLStego;

import org.apache.commons.io.IOUtils;
import org.apache.commons.io.output.ByteArrayOutputStream;
import org.apache.commons.vfs.FileContent;
import org.apache.commons.vfs.FileContentInfo;
import org.apache.commons.vfs.FileObject;
import org.apache.commons.vfs.FileSystemException;
import org.apache.commons.vfs.FileSystemManager;
import org.apache.commons.vfs.FileType;
import org.apache.commons.vfs.VFS;

/**
 * A static class that provides utility constants and methods for I/O.
 * No instances of this class can be created, only the static members
 * should be used.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.2 $, $Date: 2005/07/12 17:02:22 $, $Author: siefkes $
 */
public final class InOutUtils {

    /**
     * The line separator on the current operating system ("\n" on Unix).
     */
    public static final String LINE_SEPARATOR =
        System.getProperty("line.separator");

    /**
     * The standard block size recommended for I/O.
     */
    public static final int STANDARD_BLOCK_SIZE = 64 * 1024;

    /**
     * The standard character set defined by the HTTP standard.
     */
    public static final String STANDARD_HTTP_CHARSET = "ISO-8859-1";

    /**
     * The standard Unicode-compatible character set.
     */
    public static final String STANDARD_UNICODE_CHARSET = "UTF-8";

    /**
     * Helper method for adding the contents of file objects to a list of
     * {@link FileContent} objects.
     *
     * @param files the array of file objects
     * @param recurse whether to recursively add the children of folders and
     * other files containing nested entries
     * @param contents the objects representing the contents of the resolved
     * path names are added to this list
     * @throws IOException if an I/O error occurs, e.g. if one of the files
     * does not exist
     */
    private static void addFileContents(final FileObject[] files,
            final boolean recurse, final List<FileContent> contents)
            throws IOException {
        FileObject file;
        FileType type;

        for (int i = 0; i < files.length; i++) {
            file = files[i];
            type = file.getType();

            if (type.hasChildren()) {
                if (recurse) {
                    // recursively process children, if any
                    addFileContents(file.getChildren(), recurse, contents);
                }
            } else if (type.hasContent()) {
                // add contents
                contents.add(file.getContent());
            } else {
                // probably the file does not exist
                throw new FileNotFoundException(
                        "Invalid file (has neither content nor children): "
                        + file.getURL());
            }
        }
    }

    /**
     * Compresses data using {@linkplain GZIPOutputStream GZip} for compression.
     *
     * @param data the data to compress
     * @return a byte array containing the compressed data
     * @throws IOException if an I/O error occurs
     */
    public static byte[] compress(final byte[] data) throws IOException {
        final ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
        final GZIPOutputStream gzStream = new GZIPOutputStream(byteStream);
        try {
            gzStream.write(data);
        } finally {
            tryToClose(gzStream);
        }
        return byteStream.toByteArray();
    }

    /**
     * This method converts an array of file paths into an array of
     * {@link FileContent} objects. It uses the
     * <a href="http://jakarta.apache.org/commons/vfs/">Jakarta Commons Virtual
     * File System (VFS)</a> which allows accessing local files, URIs, ZIP
     * entries and various other resources in a uniform way.
     *
     * @param filePaths the array of file paths
     * @param recurse whether to recursively add the children of folders and
     * other files containing nested entries
     * @return an array of objects representing the contents of the resolved
     * path names
     * @throws IOException if an I/O error occurs, e.g. if one of the files
     * does not exist
     */
    public static FileContent[] listFileContents(
            final String[] filePaths, final boolean recurse)
            throws IOException {
        final FileSystemManager fsManager = VFS.getManager();
        final FileObject workingDir =
            fsManager.toFileObject(new File(userDir()));
        final FileObject[] files = new FileObject[filePaths.length];

        // resolve file paths to file objects, relative to working directory
        for (int i = 0; i < filePaths.length; i++) {
            files[i] = fsManager.resolveFile(workingDir, filePaths[i]);
        }

        // create result list and delegate to helper method
        final List<FileContent> result =
            new ArrayList<FileContent>(filePaths.length);
        addFileContents(files, recurse, result);
        return result.toArray(new FileContent[result.size()]);
    }

    /**
     * Opens a reader on a {@link FileContent} object. Uses the character set
     * specified by {@link FileContentInfo} or the default character set of
     * the current platform if no explicit charset is given.
     *
     * @param content the content to read
     * @return a reader on the given object
     * @throws FileSystemException if an error occurs while using the
     * content object
     * @throws UnsupportedEncodingException if the specified charset is not
     * available on the current platform
     */
    public static Reader openReader(final FileContent content)
    throws FileSystemException, UnsupportedEncodingException {
        final FileContentInfo info = content.getContentInfo();
        final String charset = info.getContentEncoding();
        final InputStream in = content.getInputStream();

        if (charset != null) {
            // use specified charset
            return new InputStreamReader(in, charset);
        } else {
            // use default charset
            return new InputStreamReader(in);
        }
    }

    /**
     * Reads the contents of a reader into a string. The reader is <em>not</em>
     * closed by this method.
     *
     * <p>This method should be slightly more efficient than
     * {@link org.apache.commons.io.IOUtils#toString(java.io.Reader)} since it
     * uses an unsynchronized {@link StringBuilder} instead of a synchronized
     * {@link StringBuffer} (hidden in a {@link java.io.StringWriter}).
     *
     * @param reader the reader to use
     * @return the contents of the reader
     * @throws IOException if an I/O error occurs
     */
    public static String readToString(final Reader reader) throws IOException {
        final StringBuilder result = new StringBuilder();

        final char[] buffer = new char[STANDARD_BLOCK_SIZE];
        int charsRead;

        while ((charsRead = reader.read(buffer)) > -1) {
            result.append(buffer, 0, charsRead);
        }

        return result.toString();
    }

    /**
     * Uncompresses data using {@linkplain GZIPInputStream GZip}.
     *
     * @param compressedData the data the uncompressed
     * @return the uncompressed data
     * @throws IOException if an I/O error occurs
     */
    public static byte[] uncompress(final byte[] compressedData)
    throws IOException {
        final ByteArrayInputStream byteStream =
            new ByteArrayInputStream(compressedData);
        final GZIPInputStream gzStream = new GZIPInputStream(byteStream);
        return IOUtils.toByteArray(gzStream);
    }

    /**
     * {@linkplain InputStream#read() Reads a byte} from an input stream,
     * throwing an {@link EOFException} if the end of the stream has been
     * reached.
     *
     * @param in the stream to read from
     * @return the read byte
     * @throws EOFException if the end of the stream has been reached
     * @throws IOException if the end of the stream has been reached or if any
     * other I/O error occurs
     */
    public static byte readExpectedByte(final InputStream in)
    throws EOFException, IOException {
        final int result = in.read();
        if (result < 0) {
            throw new EOFException(NLStego.RES.getString("io/unexpected-eos"));
        } else {
            return (byte) result;
        }
    }

    /**
     * Returns the user's current working directory. This is just a shortcut
     * for calling <code>System.getProperty("user.dir")</code>.
     *
     * @return the current working directory
     */
    public static String userDir() {
        return System.getProperty("user.dir");
    }

    /**
     * Returns the user's home directory. This is just a shortcut for calling
     * <code>System.getProperty("user.home")</code>.
     *
     * @return the home directory
     */
    public static String userHome() {
        return System.getProperty("user.home");
    }

    /**
     * Convenience method for closing an input stream. If the specified stream
     * is <code>null</code>, this method does nothing. Any {@link
     * java.io.IOException} thrown during closing is swallowed by this method.
     *
     * @param in the input stream to close (might be <code>null</code>)
     * @return <code>true</code> if the stream was closed successfully,
     * <code>false</code> otherwise (the stream was <code>null</code> or an
     * exception was thrown during closing)
     */
    public static boolean tryToClose(final InputStream in) {
        boolean result = false;

        if (in != null) {
            try {
                in.close();
                result = true;
            } catch (IOException ioe) {
                NLStego.LOG.warn(MessageFormat.format(NLStego.RES.getString(
                        "IO/close-instream.warn"), in, ioe.toString()));
            }
        }

        return result;
    }

    /**
     * Convenience method for closing an output stream. If the specified stream
     * is <code>null</code>, this method does nothing. Any {@link
     * java.io.IOException} thrown during closing is swallowed by this method.
     *
     * @param out the output stream to close (might be <code>null</code>)
     * @return <code>true</code> if the stream was closed successfully,
     * <code>false</code> otherwise (the stream was <code>null</code> or an
     * exception was thrown during closing)
     */
    public static boolean tryToClose(final OutputStream out) {
        boolean result = false;

        if (out != null) {
            try {
                out.close();
                result = true;
            } catch (IOException ioe) {
                NLStego.LOG.warn(MessageFormat.format(NLStego.RES.getString(
                        "IO/close-outstream.warn"), out, ioe.toString()));
            }
        }

        return result;
    }

    /**
     * Convenience method for closing a reader. If the specified reader is
     * <code>null</code>, this method does nothing. Any {@link
     * java.io.IOException} thrown during closing is swallowed by this method.
     *
     * @param reader the reader to close (might be <code>null</code>)
     * @return <code>true</code> if the reader was closed successfully,
     * <code>false</code> otherwise (reader was <code>null</code> or an
     * exception was thrown during closing)
     */
    public static boolean tryToClose(final Reader reader) {
        boolean result = false;

        if (reader != null) {
            try {
                reader.close();
                result = true;
            } catch (IOException ioe) {
                NLStego.LOG.warn(MessageFormat.format(NLStego.RES.getString(
                        "IO/close-reader.warn"), reader, ioe.toString()));
            }
        }

        return result;
    }

    /**
     * Convenience method for closing a writer. If the specified writer is
     * <code>null</code>, this method does nothing. Any {@link
     * java.io.IOException} thrown during closing is swallowed by this method.
     *
     * @param writer the writer to close (might be <code>null</code>)
     * @return <code>true</code> if the writer was closed successfully,
     * <code>false</code> otherwise (writer was <code>null</code> or an
     * exception was thrown during closing)
     */
    public static boolean tryToClose(final Writer writer) {
        boolean result = false;

        if (writer != null) {
            try {
                writer.close();
                result = true;
            } catch (IOException ioe) {
                NLStego.LOG.warn(MessageFormat.format(NLStego.RES.getString(
                        "IO/close-writer.warn"), writer, ioe.toString()));
            }
        }

        return result;
    }

    /**
     * Private constructor prevents creation of instances.
     */
    private InOutUtils() {
        super();
    }

}
