/*
 * Copyright (C) 2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.util;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import net.siefkes.nlstego.NLStego;

/**
 * A static utility class that provides methods for working with numbers.
 * No instances of this class can be created, only the static members
 * should be used.
 * 
 * @author Christian Siefkes
 * @version $Revision: 1.3 $, $Date: 2005/07/12 17:02:22 $, $Author: siefkes $
 */
public final class NumUtils {

    /**
     * Reads a number in the variable-length VInt format (as
     * <a href="http://lucene.apache.org/java/docs/fileformats.html">used by
     * Apache Lucene</a>).
     *
     * @param in the stream to read from
     * @return the decoded number
     * @throws IOException if an I/O error occurs while reading
     */
    public static long decodeVInt(final InputStream in) throws IOException {
        byte b = InOutUtils.readExpectedByte(in);
        long result = b & 0x7F;
        for (int shift = 7; !endsVInt(b); shift += 7) {
          b = InOutUtils.readExpectedByte(in);
          result |= (b & 0x7FL) << shift;
        }
        return result;
    }

    /**
     * Checks whether a byte is the last byte of a number stored in the
     * variable-length VInt format (as
     * <a href="http://lucene.apache.org/java/docs/fileformats.html">used by
     * Apache Lucene</a>). For the last byte in a VInt, the highest-order bit
     * is 0, for all other bytes in a VInt it is 1 .
     *
     * @param b the byte to check
     * @return <code>true</code> if the byte looks like the last byte in a VInt
     */
    public static boolean endsVInt(final byte b) {
        return (b & 0x80) == 0;
    }

    /**
     * Encodes a number in the variable-length VInt format (as
     * <a href="http://lucene.apache.org/java/docs/fileformats.html">used by
     * Apache Lucene</a>).
     *
     * @param num the number to write
     * @param out the stream to write the number to
     * @throws IOException if an I/O error occurs while writing
     */
    public static void encodeVInt(final long num, final OutputStream out)
    throws IOException {
        long myNum = num;
        while ((myNum & ~0x7F) != 0) {
            out.write((byte) ((myNum & 0x7f) | 0x80));
            myNum >>>= 7;
        }
        out.write((byte) myNum);
    }

    /**
     * Encodes a number in the variable-length VInt format (as
     * <a href="http://lucene.apache.org/java/docs/fileformats.html">used by
     * Apache Lucene</a>).
     *
     * @param num the number to write
     * @return a byte array encoding the given number in VInt format
     */
    public static byte[] encodeVInt(final long num) {
        final ByteArrayOutputStream out = new ByteArrayOutputStream();
        try {
            encodeVInt(num, out);
        } catch (IOException ioe) {
            // not supposed to happen when writing to byte array stream
            NLStego.LOG.warn(ioe);
        }
        return out.toByteArray();
    }


    /**
     * Private constructor prevents creation of instances.
     */
    private NumUtils() {
        super();
    }

}
