/*
 * Copyright (C) 2004-2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.util;

/**
 * Factory for creating {@link net.siefkes.nlstego.util.TextTokenizer}s of
 * different types.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.4 $, $Date: 2005/07/12 17:02:22 $, $Author: siefkes $
 */
public class TokenizerFactory {

    /**
     * Pattern string capturing whitespace and control/other characters.
     */
    public static final String WHITESPACE_CONTROL_OTHER = "[\\p{Z}\\p{C}]*";

    /**
     * Static factory method to create an instance for tokenizing alphanumeric
     * and symbol sequences and puntuation. Token types:
     *
     * <ul>
     * <li>sequences of letters and digits (incl. marks, Unicode categories
     * "L", "M", "N") -- full token is captured for
     * {@link TextTokenizer#capturedText()}</li>
     * <li>sequences of symbols ("S" category) -- nothing is captured</li>
     * <li>a single punctuation sign, possibly repeated ("P" category) --
     * nothing is captured</li>
     * </ul>
     *
     * <p>When you are only interested in words and numbers (e.g. for indexing),
     * you can use the {@link TextTokenizer#capturedText() captured text} --
     * it will contain the full token for alphanumeric sequences, it will be
     * empty for symbols and punctuation.
     *
     * <p>The whitespace pattern comprised a sequence of whitespace and
     * control/other characters ("C" and "Z" categories).
     *
     * @param text the text to tokenize
     * @return the created tokenizer
     */
    public static TextTokenizer createAlnumTokenizer(final CharSequence text) {
        final TextTokenizer result = new TextTokenizer(new String[] {
                "([\\p{L}\\p{M}\\p{N}]+)", // alphanumeric
                "\\p{S}+",        // symbols
                "(\\p{P})\\1*",   // single punctuation sign, possibly repeated
            },
            WHITESPACE_CONTROL_OTHER, // whitespace + control/other characters
            text);
        return result;
    }

    /**
     * Static factory method to create an instance for tokenizing according to
     * Unicode categories. Token types:
     *
     * <ul>
     * <li>sequences of letters (incl. marks, "L" and "M" categories) -- full
     * token is captured for {@link TextTokenizer#capturedText()}</li>
     * <li>sequences of digits (integral numbers, "N" category) -- full token is
     * captured</li>
     * <li>sequences of symbols ("S" category) -- nothing is captured</li>
     * <li>sequences of punctuation ("P" category) -- nothing is captured</li>
     * </ul>
     *
     * <p>When you are only interested in words and numbers (e.g. for indexing),
     * you can use the {@link TextTokenizer#capturedText() captured text} --
     * it will contain the full token for letter and digit sequences, it will
     * be empty for symbols and punctuation.
     *
     * <p>The whitespace pattern comprised a sequence of whitespace and
     * control/other characters ("C" and "Z" categories).
     *
     * @param text the text to tokenize
     * @return the created tokenizer
     */
    public static TextTokenizer createCategoryTokenizer(
                final CharSequence text) {
        final TextTokenizer result = new TextTokenizer(new String[] {
                "([\\p{L}\\p{M}]+)",    // letters incl. marks, fully captured
                "(\\p{N}+)",            // numbers, fully captured
                "\\p{S}+",              // symbols
                "\\p{P}+",              // punctuation
            },
            WHITESPACE_CONTROL_OTHER,   // whitespace + control/other characters
            text);
        return result;
    }

    /**
     * Static factory method to create an instance that uses the "thorough"
     * patterns listed below.
     *
     * <ul>
     * <li>sequences of letters (incl. marks)</li>
     * <li>sequences of digits (integral numbers)</li>
     * <li>sequences of math symbols</li>
     * <li>sequences of currency symbols</li>
     * <li>sequences of other symbols (modifiers and misc)</li>
     * <li>a single punctuation sign, possibly repeated</li>
     * </ul>
     *
     * <p>These patterns don't contain any useful information for
     * {@link TextTokenizer#capturedText()}.
     *
     * <p>The whitespace pattern comprised a sequence of whitespace and
     * control/other characters.
     *
     * @param text the text to tokenize
     * @return the created tokenizer
     */
    public static TextTokenizer createThoroughTokenizer(
            final CharSequence text) {
        final TextTokenizer result = new TextTokenizer(new String[] {
                "[\\p{L}\\p{M}]+",    // letters incl. marks
                "\\p{N}+",            // numbers
                "\\p{Sm}+",           // math symbols
                "\\p{Sc}+",           // currency symbols
                "[\\p{Sk}\\p{So}]+",  // modifier and other symbols
                "(\\p{P})\\1*",   // single punctuation sign, possibly repeated
            },
            WHITESPACE_CONTROL_OTHER, // whitespace + control/other characters
            text);
        return result;
    }

}
