/*
 * Copyright (C) 2003-2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.util;

import java.io.File;
import java.text.MessageFormat;
import java.text.NumberFormat;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;

import net.siefkes.nlstego.NLStego;

import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.apache.commons.lang.BooleanUtils;
import org.apache.commons.lang.builder.StandardToStringStyle;
import org.apache.commons.lang.builder.ToStringBuilder;


/**
 * A static class that provides some utility constands and methods.
 * No instances of this class can be created, only the static members
 * should be used.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.11 $, $Date: 2005/07/12 17:02:22 $, $Author: siefkes $
 */
public final class Util {

    /**
     * The line separator on the current operating system ("\n" on Unix).
     */
    public static final String LINE_SEPARATOR =
        System.getProperty("line.separator");

    /**
     * String recognized as Not-a-Number when parsing floating-point numbers:
     * {@value} (ignoring case).
     */
    public static final String NAN = "NaN";

    /**
     * Character representing the boolean value <code>true</code>: '+'.
     */
    public static final char TRUE_CHAR = '+';

    /**
     * Character representing the boolean value <code>false</code>: '-'.
     */
    public static final char FALSE_CHAR = '-';

    /**
     * The locale used for formatting data. <p>
     * Created (21.01.2002 14:49:29) by Christian Siefkes.<p>
     */
    private static final Locale FORMATTING_LOCALE = Locale.US;

    /**
     * Used to format numbers. <p>
     * Created (21.01.2002 14:48:32) by Christian Siefkes.<p>
     */
    private static final NumberFormat NUMBER_FORMATTER =
        NumberFormat.getInstance(FORMATTING_LOCALE);

    /**
     * Used to format percentages. <p>
     * Created (21.01.2002 14:36:52) by Christian Siefkes.<p>
     */
    private static final NumberFormat PERCENT_FORMATTER =
        NumberFormat.getPercentInstance(FORMATTING_LOCALE);


    /**
     * Static initializer: modify style used by {@link ToStringBuilder}.
     */
    static {
        // modify ToStringStyle: short class names, no hash code, separator: ";"
        final StandardToStringStyle tiesStyle = new StandardToStringStyle();
        tiesStyle.setUseShortClassName(true);
        tiesStyle.setUseIdentityHashCode(false);
        tiesStyle.setFieldSeparator(";");
        ToStringBuilder.setDefaultStyle(tiesStyle);
    }


    /**
     * Converts a character into a boolean primitive. {@link #TRUE_CHAR} is
     * converted to <code>true</code>; {@link #FALSE_CHAR} is
     * converted to <code>false</code>. An exception is thrown in all other
     * cases.
     *
     * @param character the character to convert
     * @return the converted boolean
     * @throws IllegalArgumentException if the character cannot be converted
     * into a boolean
     */
    public static boolean asBoolean(final char character)
            throws IllegalArgumentException {
        if (character == TRUE_CHAR) {
            return true;
        } else if (character == FALSE_CHAR) {
            return false;
        } else {
            throw new IllegalArgumentException(MessageFormat.format(
                    NLStego.RES.getString("util/cannot-convert-bool-char"),
                    character));
        }
    }

    /**
     * Converts a String into a boolean primitive.
     * Calls {@link BooleanUtils#toBooleanObject(java.lang.String)}, i.e.
     * The strings "true", "on", and "yes" are accepted for a <code>true</code>
     * value; "false", "off", and "no" are accepted for <code>false</code>;
     * case and surrounding whitespace are ignored. If the string contains a
     * single character, the result of {@link #asBoolean(char)} is returned.
     * An exception is thrown for all other strings.
     *
     * <p>If the specified String is <code>null</code>, <code>false</code>
     * is returned.
     *
     * @param str the String to check
     * @return the converted boolean
     * @throws IllegalArgumentException if the String cannot be parsed as a
     * boolean
     */
    public static boolean asBoolean(final String str)
            throws IllegalArgumentException {
        final boolean result;
        if (str != null) {
            // parse string value
            final String trimmed = str.trim();

            if (trimmed.length() == 1) {
                // single-character string
                result = asBoolean(trimmed.charAt(0));
            } else {
                final Boolean parsed = BooleanUtils.toBooleanObject(trimmed);

                if (parsed == null) {
                    // invalid string value
                    throw new IllegalArgumentException(MessageFormat.format(
                            NLStego.RES.getString(
                                "util/cannot-convert-bool-string"), trimmed));
                } else {
                    result = parsed.booleanValue();
                }
            }
        } else {
            // null object: return false
            result = false;
        }
        return result;
    }

    /**
     * Converts a String into a char primitive. The first
     * non-whitespace character String is returned. An exception is
     * thrown if the string is empty or contains only whitespace.
     *
     * <p>If the specified String is <code>null</code>,
     * {@link Character#MIN_VALUE} (the smallest character value) is returned.
     *
     * @param str the String to check
     * @return the converted char
     * @throws IndexOutOfBoundsException if the String is empty (after trimming
     * outer whitespace)
     */
    public static char asChar(final String str)
            throws IndexOutOfBoundsException {
        final char result;
        if (str != null) {
            // return first character of trimmed string value
            final String raw = str.toString().trim();
            result = raw.charAt(0);
        } else {
            // null object: return smallest character value
            result = Character.MIN_VALUE;
        }
        return result;
    }

    /**
     * Converts a String into a double primitive using the
     * {@link Double#parseDouble(java.lang.String)} method.
     * If the trimmed String to equal to {@link #NAN} (ignoring case),
     * {@link Double#NaN} is returned.
     *
     * <p>If the specified String is <code>null</code>, -1.0 is returned.
     *
     * @param str the String to check
     * @return the converted double
     * @throws NumberFormatException if the String does not contain a parsable
     * double
     */
    public static double asDouble(final String str)
            throws NumberFormatException {
        final double result;
        if (str != null) {
            final String trimmed = str.trim();

            if (NAN.equalsIgnoreCase(trimmed)) {
                // Not-a-Number
                result = Double.NaN;
            } else {
                // parse string value
                result = Double.parseDouble(trimmed);
            }
        } else {
            // null object
            result = -1.0;
        }
        return result;
    }

    /**
     * Converts a String into a float primitive
     * using the {@link Float#parseFloat(java.lang.String)} method.
     * If the trimmed String to equal to {@link #NAN} (ignoring case),
     * {@link Float#NaN} is returned.
     *
     * <p>If the specified String is <code>null</code>, -1 is returned.
     *
     * @param str the String to check
     * @return the converted float
     * @throws NumberFormatException if the String does not contain a parsable
     * float
     */
    public static float asFloat(final String str)
            throws NumberFormatException {
        final float result;
        if (str != null) {
            final String trimmed = str.trim();

            if (NAN.equalsIgnoreCase(trimmed)) {
                // Not-a-Number
                result = Float.NaN;
            } else {
                // parse string value
                result = Float.parseFloat(trimmed);
            }
        } else {
            // null object
            result = -1.0f;
        }
        return result;
    }

    /**
     * Converts a String into an integer primitive
     * using the {@link Integer#decode(java.lang.String)} method.
     * <p>If the specified String is <code>null</code>, -1 is returned.
     *
     * @param str the String to check
     * @return the converted int
     * @throws NumberFormatException if the String does not contain a parsable
     * int
     */
    public static int asInt(final String str) throws NumberFormatException {
        final int result;
        if (str != null) {
            // parse string value
            final String trimmed = str.trim();
            result = Integer.decode(trimmed).intValue();
        } else {
            // null object
            result = -1;
        }
        return result;
    }

    /**
     * Converts a String into a long primitive
     * using the {@link Long#decode(java.lang.String)} method.
     * <p>If the specified String is <code>null</code>, -1 is returned.
     *
     * @param str the String to check
     * @return the converted long
     * @throws NumberFormatException if the String does not contain a parsable
     * long
     */
    public static long asLong(final String str) throws NumberFormatException {
        final long result;
        if (str != null) {
            // parse string value
            final String trimmed = str.trim();
            result = Long.decode(trimmed).longValue();
        } else {
            // null object
            result = -1;
        }
        return result;
    }

    /**
     * Converts a String into a short primitive
     * using the {@link Short#decode(java.lang.String)} method.
     * <p>If the specified String is <code>null</code>, -1 is returned.
     *
     * @param str the String to check
     * @return the converted short
     * @throws NumberFormatException if the String does not contain a parsable
     * short
     */
    public static short asShort(final String str) throws NumberFormatException {
        final short result;
        if (str != null) {
            // parse string value
            final String trimmed = str.trim();
            result = Short.decode(trimmed).shortValue();
        } else {
            // null object
            result = -1;
        }
        return result;
    }

    /**
     * Modifies a configuration from <code>-key[=value]</code>
     * pairs in a string array. Delegates to
     * {@link #modifyConfigProperty(Configuration, String)} for each array
     * element starting with "-". Other array elements are collected and
     * returned.
     *
     * @param config the configuration to modify
     * @param args the array of strings to parse
     * @return a list of all arguments that do not start with "-"
     * @throws IllegalArgumentException if one of the
     * <code>[+|-]key[=value]</code> pairs doesn't contain a key ("=" is second
     * printable character)
     */
    public static List<String> configureFromArgs(final Configuration config,
            final String[] args) throws IllegalArgumentException {
        final List<String> otherArgs = new LinkedList<String>();

        for (int i = 0; i < args.length; i++) {
            if (args[i].startsWith("-")) {
                modifyConfigProperty(config, args[i].substring(1));
            } else {
                // adds to other arguments
                otherArgs.add(args[i]);
            }
        }

        return otherArgs;
    }

    /**
     * Formats a double, using three decimal digits. <p>
     * Created (21.01.2002 12:09:23) by Christian Siefkes.<p>
     *
     * @param number the double to format
     * @return the formatted result
     */
    public static String formatDouble(final double number) {
        return formatDouble(number, 3);
    } // method

    /**
     * Formats a double, using the specified number of
     * decimal digits. <p>
     * Created (21.01.2002 12:11:06) by Christian Siefkes.<p>
     *
     * @param number the double to format
     * @param digits how many decimal digits to use
     * @return the formatted result
     */
    public static String formatDouble(final double number, final int digits) {
        final String result;

        synchronized (NUMBER_FORMATTER) {
            // adjust digits, if necessary
            if (NUMBER_FORMATTER.getMaximumFractionDigits() != digits) {
                NUMBER_FORMATTER.setMaximumFractionDigits(digits);
            } // if

            result = NUMBER_FORMATTER.format(number);
        }

        return result;
    } // method

    /**
     * Formats a number as a percentage, using three decimal digits. <p>
     * Created (21.01.2002 12:09:23) by Christian Siefkes.<p>
     *
     * @param number the double to format
     * @return the formatted result
     */
    public static String formatPercentage(final double number) {
        return formatPercentage(number, 3);
    } // method

    /**
     * Formats a number as a percentage, using the specified number of
     * decimal digits. <p>
     * Created (21.01.2002 12:11:06) by Christian Siefkes.<p>
     *
     * @param number the double to format
     * @param digits how many decimal digits to use
     * @return the formatted result
     */
    public static String formatPercentage(final double number,
            final int digits) {
        final String result;

        synchronized (PERCENT_FORMATTER) {
            // adjust digits, if necessary
            if (PERCENT_FORMATTER.getMaximumFractionDigits() != digits) {
                PERCENT_FORMATTER.setMaximumFractionDigits(digits);
            } // if

            result = PERCENT_FORMATTER.format(number);
        }
        return result;
    } // method

    /**
     * Loads a configuration in {@link PropertiesConfiguration} from the
     * <code><em>baseName</em>.cfg</code> in the user's home directory, if it
     * exists and is a readable file. Otherwise an empty configuration will
     * be returned. The file is assumed to use the platform's default
     * character set.
     *
     * @param baseName the base name of the configuration
     * @return the loaded configuration
     * @throws ConfigurationException if an error occurs while loading the
     * configuraiton
     */
    public static Configuration loadConfiguration(final String baseName)
    throws ConfigurationException {
        final PropertiesConfiguration config;
        final File file = new File(System.getProperty("user.home")
                + File.separator + baseName + ".cfg");

        if (file.canRead() && file.isFile()) {
            // read config from file 
            config = new PropertiesConfiguration(file);
        } else {
            // create empty config
            config = new PropertiesConfiguration();
        }

        return config;
    }

    /**
     * Modifies a configuration property, parsing a <code>key[=value]</code>
     * pair. Both key and value are trimmed. {@link Boolean#TRUE} is used as
     * default value if the <code>=value</code> part is omitted.
     * If "=" is the last character (empty value), the property is removed from
     * the configuration
     * ({@link Configuration#clearProperty(java.lang.String)}).
     *
     * @param config the configuration to modify
     * @param keyValue the <code>key=value</code> pair to parse
     * @throws IllegalArgumentException if <code>keyValue</code> doesn't contain
     * a key ("=" is first character)
     */
    public static void modifyConfigProperty(final Configuration config,
            final String keyValue) throws IllegalArgumentException {
        // parse string
        final String key;
        final Object value;
        final int equalIndex = keyValue.indexOf('=');

        if (equalIndex > 0) {
            key = keyValue.substring(0, equalIndex).trim();
            value = keyValue.substring(equalIndex + 1).trim();
        } else if (equalIndex < 0) {
            // boolean "true" can be omitted (default value)
            key = keyValue.trim();
            // specify String instead of boolean to avoid problems during save
            value = "true";
        } else {
            // '=' is first character
            throw new IllegalArgumentException(MessageFormat.format(
                    NLStego.RES.getString("util/missing-key"), keyValue));
        }

        // empty string means: delete value
        if ("".equals(value)) {
            config.clearProperty(key);
        } else {
            config.setProperty(key, value);
        }
    }

    /**
     * Private constructor prevents creation of instances.
     */
    private Util() {
        super();
    }

}
