/*
 * Copyright (C) 2005 Christian Siefkes <christian@siefkes.net>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package net.siefkes.nlstego.util;

import java.io.PrintStream;
import java.util.Date;

import org.apache.commons.logging.impl.SimpleLog;

/**
 * A extensioin of {@link org.apache.commons.logging.impl.SimpleLog} that works
 * around a small bug in the superclass: it avoids printing the name and message
 * of throwables twice. It is also slightly more efficient by directly writing
 * to a {@link java.io.PrintStream} (by default, {@link java.lang.System#err})
 * instead of creating intermediate objects.
 *
 * @author Christian Siefkes
 * @version $Revision: 1.7 $, $Date: 2005/07/12 17:02:22 $, $Author: siefkes $
 */
public class VerySimpleLog extends SimpleLog {
 
    /**
     * The short name of this simple log instance. We have to copy this field
     * from the superclass since it is private.
     */
    private String shortLogName = null;

    /**
     * The stream to print messages to.
     */
    private final PrintStream out;


    /**
     * Construct a named instance of this Logger. All output will be printed to
     * {@link System#err}.
     *
     * @param name Name of the logger to be constructed
     */
    public VerySimpleLog(final String name) {
        this(name, System.err);
    }

   /**
     * Construct a named instance of this Logger.
     *
     * @param name Name of the logger to be constructed
     * @param outStream the stream to print output to
     */
    public VerySimpleLog(final String name, final PrintStream outStream) {
        super(name);
        out = outStream;
        //setLevel(LOG_LEVEL_DEBUG); // for debugging
    }


    /**
     * {@inheritDoc}
     */
    protected void log(final int type, final Object message,
            final Throwable t) {
        // synchronize on the stream to avoid overlaps
        synchronized (out) {
            // Append date-time if so configured
            if (showDateTime) {
                out.print(dateFormatter.format(new Date()) + " ");
            }

            // Append a readable representation of the log level
            switch (type) {
                case SimpleLog.LOG_LEVEL_TRACE: out.print("[TRACE] "); break;
                case SimpleLog.LOG_LEVEL_DEBUG: out.print("[DEBUG] "); break;
                case SimpleLog.LOG_LEVEL_INFO:  out.print("[INFO] ");  break;
                case SimpleLog.LOG_LEVEL_WARN:  out.print("[WARN] ");  break;
                case SimpleLog.LOG_LEVEL_ERROR: out.print("[ERROR] "); break;
                case SimpleLog.LOG_LEVEL_FATAL: out.print("[FATAL] "); break;
                default: throw new IllegalArgumentException(
                        "Unknown log level: " + type);
            }

            // Append the name of the log instance if so configured
            if (showShortName) {
                if (shortLogName == null) {
                    // Cut all but the last component of name for both styles
                    shortLogName =
                        logName.substring(logName.lastIndexOf(".") + 1);
                    shortLogName = shortLogName.substring(
                            shortLogName.lastIndexOf("/") + 1);
                }
                out.print(String.valueOf(shortLogName) + " - ");
            } else if (showLogName) {
                out.print(String.valueOf(logName) + " - ");
            }

            // Append the message
            out.print(String.valueOf(message));

            // Append stack trace if not null
            if (t != null) {
                out.print(" - ");
                // this method generally prints t.toString() in the first line
                // so it's not necessary to invoke that method explicitly
                t.printStackTrace(out);
            } else {
                // terminate line
                out.println();
            }

            // flush the stream (in case autoFlush is not enabled)
            out.flush();
        }
    }

}
